# Copyright (C) 2017 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""vdo.py: Module to manipulate functionality (deduplication and compression) provided by VDO."""

__author__ = "Jakub Krysl"
__copyright__ = "Copyright (c) 2017 Red Hat, Inc. All rights reserved."

import os
import stat

from libsan import _print
from libsan.host.cli_tools import (
    FailedCheckExceptionError,
    Wrapper,
    WrongArgumentExceptionError,
    WrongCommandExceptionError,
)
from libsan.host.cmdline import run


class VDO(Wrapper):
    def __init__(self, disable_check=False):
        self.disable_check = disable_check

        self.commands = {
            "create": "create",
            "remove": "remove",
            "start": "start",
            "stop": "stop",
            "activate": "activate",
            "deactivate": "deactivate",
            "status": "status",
            "list": "list",
            "modify": "modify",
            "change_write_policy": "changeWritePolicy",
            "enable_deduplication": "enableDeduplication",
            "disable_deduplication": "disableDeduplication",
            "enable_compression": "enableCompression",
            "disable_compression": "disableCompression",
            "grow_logical": "growLogical",
            "grow_physical": "growPhysical",
            "print_config_file": "printConfigFile",
        }
        self.commands["all"] = list(self.commands.keys())

        self.arguments = {
            "all": [self.commands["all"], " --all"],
            "conf_file": [self.commands["all"], " --confFile="],
            "log_file": [self.commands["all"], " --logfile="],
            "name": [self.commands["all"], " --name="],
            "no_run": [self.commands["all"], " --noRun"],
            "verbose": [self.commands["all"], " --verbose"],
            "activate": [["create"], " --activate="],
            "compression": [["create"], " --compression="],
            "deduplication": [["create"], " --deduplication="],
            "device": [["create"], " --device="],
            "emulate512": [["create"], " --emulate512="],
            "index_mem": [["create"], " --indexMem="],
            "sparse_index": [["create"], " --sparseIndex="],
            "logical_size": [["create", "grow_logical"], " --vdoLogicalSize="],
            "log_level": [["create"], " --vdoLogLevel="],
            "slab_size": [["create"], " --vdoSlabSize="],
            "block_map_cache_size": [["create", "modify"], " --blockMapCacheSize="],
            "block_map_period": [["create", "modify"], " --blockMapPeriod="],
            "max_discard_size": [["create", "modify"], " --maxDiscardSize="],
            "ack_threads": [["create", "modify"], " --vdoAckThreads="],
            "bio_rotation_interval": [["create", "modify"], " --vdoBioRotationInterval="],
            "bio_threads": [["create", "modify"], " --vdoBioThreads="],
            "cpu_threads": [["create", "modify"], " --vdoCpuThreads="],
            "hash_zone_threads": [["create", "modify"], " --vdoHashZoneThreads="],
            "logical_threads": [["create", "modify"], " --vdoLogicalThreads="],
            "physical_threads": [["create", "modify"], " --vdoPhysicalThreads="],
            "write_policy": [["create", "modify", "change_write_policy"], " --writePolicy="],
            "force_rebuild": [["start"], " --forceRebuild"],
            "force": [["stop", "remove", "create"], " --force"],
        }

        Wrapper.__init__(self, self.commands, self.arguments, self.disable_check)

    @staticmethod
    def _check_size_format(size, return_size=False):
        # check if requested size format is in supported formats and the rest is numbers
        # FIXME: Is KiB and KB valid too?
        size = size.strip("'")
        try:
            if size[-3:] in ["KiB", "MiB", "GiB", "TiB"] and isinstance(int(size[:-3]), int):
                if return_size:
                    return True, [size[:-3], size[-3:-2]]
                return True
            elif size[-2:] in ["KB", "MB", "GB", "TB"] and isinstance(int(size[:-2]), int):
                if return_size:
                    return True, [size[:-2], size[-2:-1]]
                return True
            elif size[-1:].upper() in ["K", "M", "G", "T"] and isinstance(int(size[:-1]), int):
                if return_size:
                    return True, [size[:-1], size[-1:]]
                return True
            elif int(size):
                if return_size:
                    # default size is megabytes
                    return True, [size, "M"]
                return True
        except ValueError:
            pass
        return False, []

    @staticmethod
    def _is_positive_int(value):
        try:
            port = int(value)
            if port < 1:
                raise ValueError  # noqaTRY301
        except ValueError:
            return False
        return True

    def _check(self, cmd):
        if self.disable_check:
            # Do not check if checking is disabled
            return True

        if self._get_arg("all") in cmd and self._get_arg("name") in cmd:
            _print("WARN: Use either 'name' or 'all', not both.")
            raise FailedCheckExceptionError()

        if self._get_arg("conf_file") in cmd:
            _file = self._get_value(cmd, self._get_arg("conf_file"))
            if not os.path.isfile(_file):
                _print("WARN: Config file %s is not a regular file." % _file)
                raise FailedCheckExceptionError(self._get_arg("conf_file"))

        if self._get_arg("log_file") in cmd:
            _file = self._get_value(cmd, self._get_arg("log_file"))
            if not os.path.isfile(_file) and stat.S_ISBLK(os.stat(_file).st_mode):
                _print("WARN: Path %s exists and is not a regular file." % _file)
                raise FailedCheckExceptionError(self._get_arg("log_file"))

        if self._get_arg("name") in cmd:
            # FIXME: Check if VDO already exists
            pass

        for arg in ["activate", "compression", "deduplication", "emulate512", "sparse_index"]:
            if self._get_arg(arg) in cmd:
                _value = self._get_value(cmd, self._get_arg(arg))
                if _value not in ["enabled", "disabled"]:
                    _print("WARN: %s value must be either 'enabled' or 'disabled'." % arg)
                    raise FailedCheckExceptionError(self._get_arg(arg))

        for arg in ["logical_size", "slab_size", "block_map_cache_size", "max_discard_size"]:
            if self._get_arg(arg) in cmd:
                _value = self._get_value(cmd, self._get_arg(arg))
                ret, _ = self._check_size_format(_value, return_size=True)
                if not ret:
                    _print("WARN: VDO {} value {} is in unknown format.".format(" ".join(arg.split("_")), _value))
                    raise FailedCheckExceptionError(self._get_arg(arg))
                if arg == "slab_size":  # noqaSIM114
                    pass
                    # FIXME: Check if size is power of 2 between 128M and 32G
                elif arg == "block_map_cache_size":
                    pass
                    # FIXME: Check if size is multiple of 4096

        if self._get_arg("index_mem") in cmd:
            _value = self._get_value(cmd, self._get_arg("index_mem"), return_type=float)
            if not (_value in [0, 0.25, 0.5, 0.75] or self._is_positive_int(_value)):
                _print("WARN: Albireo mem value %s is not a 0, 0.25, 0.5, 0.75 or positive int." % _value)
                raise FailedCheckExceptionError(self._get_arg("index_mem"))

        if self._get_arg("log_level") in cmd:
            _value = self._get_value(cmd, self._get_arg("log_level"))
            possible_values = ["critical", "error", "warning", "notice", "info", "debug"]
            if _value not in possible_values:
                _print("WARN: Unknown vdo log level value, must be one of %s." % possible_values)
                raise FailedCheckExceptionError(self._get_arg("log_level"))

        if self._get_arg("device") in cmd:
            _value = self._get_value(cmd, self._get_arg("device"))
            # FIXME: Check if device exists

        if self._get_arg("block_map_period") in cmd:
            _value = self._get_value(cmd, self._get_arg("block_map_period"))
            if not self._is_positive_int(_value):
                _print("WARN: Block map period value must be a positive integer.")
                raise FailedCheckExceptionError(self._get_arg("block_map_period"))
            # FIXME: Can this be higher than 16380?

        for arg in [
            "ack_threads",
            "bio_rotation_interval",
            "bio_threads",
            "cpu_threads",
            "hash_zone_threads",
            "logical_threads",
            "physical_threads",
        ]:
            if self._get_arg(arg) in cmd:
                _value = self._get_value(cmd, self._get_arg(arg))
                if not self._is_positive_int(_value):
                    _print("WARN: VDO %s value must be a positive integer." % " ".join(arg.split("_")))
                    raise FailedCheckExceptionError(self._get_arg(arg))
                    # FIXME: Is 0 valid?

        if self._get_arg("write_policy") in cmd:
            _value = self._get_value(cmd, self._get_arg("write_policy"))
            if _value not in ["sync", "async"]:
                _print("WARN: VDO read cache value must be either 'sync' or 'async'.")
                raise FailedCheckExceptionError(self._get_arg("write_policy"))

        if self._get_arg("force_rebuild") in cmd and self._get_arg("upgrade") in cmd:
            _print("WARN: Cannot use both force_rebuild and upgrade when starting VDO volume.")
            raise FailedCheckExceptionError()

        return True

    def _run(self, cmd, verbosity=True, return_output=False, **kwargs):
        # Constructs the command to run and runs it

        ret_fail = False
        if return_output:
            ret_fail = (False, None)

        try:
            command = self._add_command(cmd)
            command = self._add_arguments(command, **kwargs)

        except WrongCommandExceptionError as e:
            _print("WARN: Given command '%s' is not allowed in this VDO version." % e.command)
            return ret_fail
        except WrongArgumentExceptionError as e:
            message = "WARN: Given argument '%s' is not allowed for given command." % e.argument
            if e.command:
                message = message[:-1] + " '" + e.command + "'."
            if e.arguments:
                message += "\nPlease use only these: %s." % ", ".join(e.arguments)
            _print(message)
            return ret_fail

        cmd = "vdo " + command

        try:
            self._check(cmd)
        except WrongArgumentExceptionError:
            pass
        except FailedCheckExceptionError as e:
            _print("WARN: Failed checking on argument %s" % e.argument)
            return ret_fail

        if return_output:
            ret, data = run(cmd, verbose=verbosity, return_output=True)
            if ret != 0:
                _print("WARN: Running command: '%s' failed. Return with output." % cmd)
            return ret, data
        ret = run(cmd, verbose=verbosity)
        if ret != 0:
            _print("WARN: Running command: '%s' failed." % cmd)
        return ret

    @staticmethod
    def help():
        if run("vdo --help", verbose=True) != 0:
            _print("WARN: Running command: 'vdo --help' failed.")
            return False
        return True

    def create(self, **kwargs):
        return self._run("create", **kwargs)

    def remove(self, force=True, **kwargs):
        return self._run("remove", force=force, **kwargs)

    def start(self, **kwargs):
        return self._run("start", **kwargs)

    def stop(self, force=True, **kwargs):
        return self._run("stop", force=force, **kwargs)

    def activate(self, **kwargs):
        return self._run("activate", **kwargs)

    def deactivate(self, **kwargs):
        return self._run("deactivate", **kwargs)

    def status(self, **kwargs):
        return self._run("status", **kwargs)

    def list(self, **kwargs):
        return self._run("list", **kwargs)

    def modify(self, **kwargs):
        return self._run("modify", **kwargs)

    def change_write_policy(self, **kwargs):
        return self._run("change_write_policy", **kwargs)

    def deduplication(self, enable=True, **kwargs):
        ret = self._run("enable_deduplication", **kwargs) if enable else self._run("disable_deduplication", **kwargs)
        return ret

    def compression(self, enable=True, **kwargs):
        ret = self._run("enable_compression", **kwargs) if enable else self._run("disable_compression", **kwargs)
        return ret

    def grow(self, grow_type=None, **kwargs):
        if grow_type.upper() not in ["LOGICAL", "PHYSICAL"]:
            _print("WARN: Please specify either 'logical' or 'physical' type for growing VDO.")
            if kwargs["return_output"]:
                return False, None
            return False

        if grow_type.upper() == "LOGICAL":
            ret = self._run("grow_logical", **kwargs)
        else:
            ret = self._run("grow_physical", **kwargs)
        return ret

    def print_config_file(self, **kwargs):
        return self._run("print_config_file", **kwargs)


class VDOStats:
    def __init__(self, disable_check=False):
        self.disable_check = disable_check
        self.command = "vdostats"
        self.arguments = {
            "help": " --help",
            "all": " --all",
            "human_readable": " --human-readable",
            "si": " --si",
            "verbose": " --verbose",
            "version": " --version",
        }

    def _get_arg(self, name):
        return self.arguments[name]

    def _get_possible_arguments(self):
        # Returns possible arguments
        return list(self.arguments.keys())

    def _add_argument(self, arg, command):
        # Checks if given argument is allowed and adds it to cmd string
        if arg not in self.arguments:
            return None
        argument = self._get_arg(arg)
        command += argument
        return command

    def _add_arguments(self, cmd, **kwargs):
        command = cmd
        for kwarg in kwargs:
            command = self._add_argument(kwarg, command)
            if command is None:
                args = self._get_possible_arguments()
                _print(f"WARN: Unknown argument '{kwarg}', please use only these: {args}.")
                return None
        return command

    def _check(self, cmd):
        if self.disable_check:
            # Do not check if checking is disabled
            return True

        # check if specified devices are block devices
        for block in cmd.split():
            if block not in list(self.arguments.values()) and block != self.command:  # noqa: SIM102
                if os.path.exists(block) and not stat.S_ISBLK(os.stat(block).st_mode):
                    _print("WARN: Device %s is not a block device." % block)
                    return False

        return True

    def _run(self, **kwargs):
        # Constructs the command to run and runs it
        cmd = self.command

        if "devices" in kwargs:
            devices = kwargs.pop("devices")
            if isinstance(devices, list):
                for device in devices:
                    cmd += " " + str(device)
            else:
                cmd += " " + str(devices)

        cmd = self._add_arguments(cmd, **kwargs)
        if cmd is None:
            return False

        if not self._check(cmd):
            # Requested command did not pass checking, reason was already written by _check()
            return False

        if run(cmd, verbose=True) != 0:
            _print("WARN: Running command: '%s' failed." % cmd)
            return False
        return True

    def help(self):
        if not self._run(help=True):
            return False
        return True

    def version(self):
        if not self._run(version=True):
            return False
        return True

    def stats(self, **kwargs):
        if not self._run(**kwargs):
            return False
        return True



#!/usr/bin/python
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""
Tool to perform basic commands on SAN
"""

import argparse
import json
import sys

import libsan.host.dt
import libsan.host.fcoe
import libsan.host.fio
import libsan.host.linux
import libsan.host.lio
import libsan.host.net
import libsan.host.scsi
import libsan.sanmgmt
from libsan.misc.time import get_time, time_2_sec

OBJ_SANMGMT = None


def _show_san_conf():
    """
    Show SAN config file
    """
    san_conf = libsan.sanmgmt.load_san_conf()
    print(json.dumps(san_conf, indent=4))


def _show_host_cfg():
    """
    Show SAN configuration of the host
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    nics = libsan.host.net.get_nics()
    macs = []
    if nics:
        for nic in nics:
            macs.append(libsan.host.net.get_mac_of_nic(nic))
    if macs:
        print("Host has the following MACs:")
        for mac in macs:
            sw_port = "Not configured on SAN top"
            connected_sw = OBJ_SANMGMT.get_sw_self(mac)
            if connected_sw:
                sw_port = connected_sw["port_of"][mac]
            print(f"\t{mac} - SW port: {sw_port}")

    iscsi_macs = OBJ_SANMGMT.macs()
    if iscsi_macs:
        print("Host has the following iSCSI MACs:")
        for mac in iscsi_macs:
            sw_port = "Not configured on SAN top"
            connected_sw = OBJ_SANMGMT.get_sw_self(mac)
            if connected_sw:
                sw_port = connected_sw["port_of"][mac]
            print(f"\t{mac} - SW port: {sw_port}")

    h_wwpns = OBJ_SANMGMT.h_wwpns()
    if h_wwpns:
        print("Host has the following h_wwpns configured:")
        for wwpn in h_wwpns:
            sw_port = "Not configured on SAN top"
            connected_sw = OBJ_SANMGMT.get_sw_self(wwpn)
            if connected_sw:
                sw_port = connected_sw["port_of"][wwpn]
            print(f"\t{wwpn} - SW port: {sw_port}")

    t_wwpns = OBJ_SANMGMT.t_wwpns()
    if t_wwpns:
        print("Host has the following t_wwpns configured:")
        for wwpn in t_wwpns:
            connected_sw = OBJ_SANMGMT.get_sw_self(wwpn)
            if connected_sw:
                sw_port = connected_sw["port_of"][wwpn]
                print(f"\t{wwpn} - SW port: {sw_port}")


def _check_ports_ready():
    """
    Check if all ports need for test are up
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)
    return OBJ_SANMGMT.check_ports_ready()


def _rcsn_enable(port_addr):
    """
    Enable RCSN
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr:
        print("FAIL: rcsn_enable() - requires port_addr as parameter")
        return False

    connected_sw = OBJ_SANMGMT.get_sw_self(port_addr)
    if not connected_sw:
        print("FAIL: %s is not managed by libsan" % port_addr)
        return False
    return OBJ_SANMGMT.rcsn_enable(addr=port_addr)


def _rcsn_disable(port_addr):
    """
    Disable RCSN
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr:
        print("FAIL: rcsn_disable() - requires port_addr as parameter")
        return False

    connected_sw = OBJ_SANMGMT.get_sw_self(port_addr)
    if not connected_sw:
        print("FAIL: %s is not managed by libsan" % port_addr)
        return False
    return OBJ_SANMGMT.rcsn_disable(addr=port_addr)


def _port_disconnect(port_addr):
    """
    Disable switch port
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr:
        print("FAIL: port_disconnect() - requires port_addr as parameter")
        return False

    connected_sw = OBJ_SANMGMT.get_sw_self(port_addr)
    if not connected_sw:
        print("FAIL: %s is not managed by libsan" % port_addr)
        return False
    return OBJ_SANMGMT.link_trigger(action="DOWN", addr=port_addr)


def _port_oscillate(port_addr):
    """
    Oscilate switch port
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr:
        print("FAIL: port_oscillate() - requires port_addr as parameter")
        return False

    connected_sw = OBJ_SANMGMT.get_sw_self(port_addr)
    if not connected_sw:
        print("FAIL: %s is not managed by libsan" % port_addr)
        return False
    return OBJ_SANMGMT.link_oscillate(addr=port_addr)


def _phy_port_disconnect(port_addr):
    """
    Disable phy switch port
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr:
        print("FAIL: phy_port_disconnect() - requires port_addr as parameter")
        return False

    connected_sw = OBJ_SANMGMT.get_physw_self(port_addr)
    if not connected_sw:
        print("FAIL: %s is not managed by libsan" % port_addr)
        return False
    return OBJ_SANMGMT.phy_link_trigger(action="DOWN", addr=port_addr)


def _port_connect(port_addr):
    """
    Enable switch port
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr:
        print("FAIL: port_connect() - requires port_addr as parameter")
        return False

    connected_sw = OBJ_SANMGMT.get_sw_self(port_addr)
    if not connected_sw:
        print("FAIL: %s is not managed by libsan" % port_addr)
        return False
    return OBJ_SANMGMT.link_trigger(action="UP", addr=port_addr)


def _link_torture(port_addr, runtime, interval):
    """
    Bring switch port down and up for specific time
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr or not runtime:
        print("FAIL: link_torture() - requires port_addr, runtime as parameteres")
        return False

    if interval is None:
        interval = 30

    interval = time_2_sec(interval)
    if interval is None:
        print("FAIL: link_torture() - Could not convert interval to seconds")
        return False

    time_in_sec = time_2_sec(runtime)
    if not time_in_sec:
        return True

    current_time = get_time(in_seconds=True)
    if not current_time:
        print("FAIL: link_torture() - Could not get current time")
        return False

    end_time = current_time + time_in_sec

    error = False
    while current_time < end_time:
        if not _port_disconnect(port_addr):
            error = True
        print("INFO: sleeping for %d seconds" % interval)
        libsan.host.linux.sleep(interval)
        if not _port_connect(port_addr):
            error = True
        if error:
            break
        print("INFO: sleeping for %d seconds" % interval)
        libsan.host.linux.sleep(interval)
        current_time = get_time(in_seconds=True)

    if error:
        # Try to bring port up
        print("Trying to bring %s back up" % port_addr)
        _port_connect(port_addr)
        return False

    return True


def _phy_port_connect(port_addr):
    """
    Enable phy switch port
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr:
        print("FAIL: phy_port_connect() - requires port_addr as parameter")
        return False

    connected_sw = OBJ_SANMGMT.get_physw_self(port_addr)
    if not connected_sw:
        print("FAIL: %s is not managed by libsan" % port_addr)
        return False
    return OBJ_SANMGMT.phy_link_trigger(action="UP", addr=port_addr)


def _phy_port_oscillate(port_addr):
    """
    Oscilate phy switch port
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not port_addr:
        print("FAIL: phy_port_oscillate() - requires port_addr as parameter")
        return False

    connected_sw = OBJ_SANMGMT.get_physw_self(port_addr)
    if not connected_sw:
        print("FAIL: %s is not managed by libsan" % port_addr)
        return False
    return OBJ_SANMGMT.phy_link_oscillate(addr=port_addr)


def _lun_create(lun_size, lun_name=None):
    """
    Create a LUN on array
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not lun_size:
        print("FAIL: lun_create() - requires size as parameter")
        return None

    new_lun_name = OBJ_SANMGMT.lun_create(lun_size, lun_name=lun_name)
    if new_lun_name:
        print("INFO: created LUN %s" % new_lun_name)
        return new_lun_name
    print("FAIL: Could not create LUN")
    return None


def _lun_create_and_map(lun_size, lun_name=None):
    """
    Create a LUN to array and map it to server
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: lun_create_and_map() - Could not create LibSAN object")
            sys.exit(1)

    if not lun_size:
        print("FAIL: lun_create_and_map() - requires size as parameter")
        return None

    new_lun_name = OBJ_SANMGMT.lun_create_and_map(lun_size, lun_name=lun_name, rescan=True)
    if new_lun_name:
        print("INFO: created and mapped LUN %s" % new_lun_name)
        return new_lun_name
    print("FAIL: Could not create LUN")
    return None


def _lun_remove(lun_name):
    """
    Remove LUN from array
    """
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not lun_name:
        print("FAIL: lun_remove() - requires lun_name as parameter")
        return False

    error = 0
    names = lun_name.split(",")
    for name in names:
        if not OBJ_SANMGMT.lun_remove(name):
            print("FAIL: Could not delete %s" % name)
            error += 1
    if error > 0:
        return False
    return True


def _lun_map(lun_name, t_addr=None, i_addr=None):
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not lun_name:
        print("FAIL: lun_map() - requires lun_name as parameter")
        return False

    return OBJ_SANMGMT.lun_map(lun_name, t_addr=t_addr, i_addr=i_addr, rescan=True)


def _lun_unmap(lun_name, t_addr=None, i_addr=None):
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not lun_name:
        print("FAIL: lun_unmap() - requires lun_name as parameter")
        return False

    return OBJ_SANMGMT.lun_unmap(lun_name, t_addr=t_addr, i_addr=i_addr)


def _lun_info(lun_name):
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    if not lun_name:
        print("FAIL: lun_info() - requires lun_name as parameter")
        return False

    info = OBJ_SANMGMT.lun_info(lun_name)
    if info:
        print(info)
        return True
    print("FAIL: Could not get info for %s" % lun_name)
    return False


def _lun_show():
    global OBJ_SANMGMT
    if not OBJ_SANMGMT:
        # Initialize SAN obj based on all information configured on /etc/san_top.cfg
        # based on HW information
        OBJ_SANMGMT = libsan.sanmgmt.create_sanmgmt_for_mpath()
        if not OBJ_SANMGMT:
            print("FAIL: Could not create LibSAN object")
            sys.exit(1)

    luns = OBJ_SANMGMT.lun_query()
    if not luns:
        return False
    for lun_name in luns:
        print("%s" % lun_name)
    return True


def _remove_device_wwid(wwid):
    return libsan.host.linux.remove_device_wwid(wwid)


def _rescan_host(hostid):
    return libsan.host.scsi.rescan_host(hostid)


def _dt_stress(of, log=None, thread=None, runtime=None):
    return libsan.host.dt.dt_stress(of, log=log, thread=thread, time=runtime)


def _dt_stress_background(of, log=None, thread=None, runtime=None):
    if libsan.host.dt.dt_stress_background(of, log=log, thread=thread, time=runtime):
        return True
    return False


def _fio_stress(of, threads=None, runtime=None, verify=None):
    ret, output = libsan.host.fio.fio_stress(
        of, threads=threads, runtime=runtime, verify=verify, return_output=True, time_based=None
    )
    if output:
        print(output)
    return ret


def _fio_stress_background(of, threads=None, runtime=None, verify=None):
    if libsan.host.fio.fio_stress_background(of, threads=threads, runtime=runtime, verify=verify):
        return True
    return False


arg_lun_name = {"name": "--name -n", "help": "Lun name", "dest": "lun_name"}
arg_lun_size = {
    "name": "--size -s",
    "help": "Lun Size. Eg. 2MiB, 2GiB,<F12> 2TiB",
    "dest": "lun_size",
    "metavar": "size",
    "required": True,
}
arg_init_addr = {
    "name": "--i_addr -i",
    "required": False,
    "dest": "i_addr",
    "metavar": "init_addr",
    "help": "Initiator address. IQN/WWPN",
}
arg_tgt_addr = {
    "name": "--t_addr -t",
    "required": False,
    "dest": "t_addr",
    "metavar": "target_addr",
    "help": "Target address. IQN/WWPN",
}
arg_port_addr = {
    "name": "--addr -a",
    "required": True,
    "dest": "port_addr",
    "metavar": "addr",
    "help": "WWPN/MAC of port",
}
arg_wwid = {"name": "--wwid -w", "required": True, "dest": "wwid", "metavar": "wwid", "help": "SCSI device WWID"}
arg_hostid = {"name": "--hostid -H", "required": False, "dest": "hostid", "metavar": "hostid", "help": "Host ID"}
arg_of = {
    "name": "--of -o",
    "required": True,
    "dest": "of",
    "metavar": "output",
    "help": "output, can be a device or file",
}
arg_runtime = {
    "name": "--time -t",
    "required": True,
    "dest": "runtime",
    "metavar": "time",
    "help": "For how long it should run",
}
arg_log = {"name": "--log -l", "required": False, "dest": "log", "metavar": "log", "help": "File to store log"}
arg_threads = {
    "name": "--thread -T",
    "required": False,
    "dest": "threads",
    "metavar": "threads",
    "help": "Number of threads",
}
arg_fio_verify = {
    "name": "--verify -V",
    "required": False,
    "dest": "verify",
    "metavar": "type",
    "choices": ["crc32c", "md5"],
}

cmds = (
    {
        "name": "show_san_conf",
        "help": "Show san_top config file",
        "func": _show_san_conf,
    },
    {
        "name": "show_host_config",
        "help": "Show information about host configuration",
        "func": _show_host_cfg,
    },
    {
        "name": "setup_soft_fcoe",
        "help": "Configure Soft FCoE",
        "func": libsan.host.fcoe.setup_soft_fcoe,
    },
    {
        "name": "create_basic_iscsi_target",
        "help": "Setup a basic iSCSI target",
        "func": libsan.host.lio.lio_setup_iscsi_target,
    },
    {
        "name": "lun_create",
        "help": "Create new LUN",
        "func": _lun_create,
        "args": [
            dict(arg_lun_name),
            dict(arg_lun_size),
        ],
    },
    {
        "name": "lun_create_and_map",
        "help": "Create new LUN and map it to target",
        "func": _lun_create_and_map,
        "args": [
            dict(arg_lun_name),
            dict(arg_lun_size),
        ],
    },
    {
        "name": "lun_remove",
        "help": "Remove LUN",
        "func": _lun_remove,
        "args": [
            dict(arg_lun_name, required=True),
        ],
    },
    {
        "name": "lun_map",
        "help": "Map a LUN to initiator/target",
        "func": _lun_map,
        "args": [
            dict(arg_lun_name, required=True),
            dict(arg_init_addr),
            dict(arg_tgt_addr),
        ],
    },
    {
        "name": "lun_unmap",
        "help": "Unmap a LUN from initiator/target",
        "func": _lun_unmap,
        "args": [
            dict(arg_lun_name, required=True),
            dict(arg_init_addr),
            dict(arg_tgt_addr),
        ],
    },
    {
        "name": "lun_info",
        "help": "Show information about specific LUN",
        "func": _lun_info,
        "args": [
            dict(arg_lun_name, required=True),
        ],
    },
    {
        "name": "lun_show",
        "help": "List all LUNs from the connected array",
        "func": _lun_show,
    },
    {
        "name": "check_ports_ready",
        "help": "Check if switch ports are ready",
        "func": _check_ports_ready,
    },
    {
        "name": "rcsn_enable",
        "help": "Enable RCSN on Switch",
        "func": _rcsn_enable,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "rcsn_disable",
        "help": "Disable RCSN on Switch",
        "func": _rcsn_disable,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "port_connect",
        "help": "Bring switch port online",
        "func": _port_connect,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "port_disconnect",
        "help": "Bring switch port offline",
        "func": _port_disconnect,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "port_oscillate",
        "help": "Oscillate switch port online/offline",
        "func": _port_oscillate,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "link_torture",
        "help": "Bring switch port down and up for specific time",
        "func": _link_torture,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "phy_port_connect",
        "help": "Bring phy switch port online",
        "func": _phy_port_connect,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "phy_port_disconnect",
        "help": "Bring phy switch port offline",
        "func": _phy_port_disconnect,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "phy_port_oscillate",
        "help": "Oscillate phy switch port online/offline",
        "func": _phy_port_oscillate,
        "args": [
            dict(arg_port_addr),
        ],
    },
    {
        "name": "remove_device_wwid",
        "help": "Remove an SCSI device based on it's WWID",
        "func": _remove_device_wwid,
        "args": [
            dict(arg_wwid),
        ],
    },
    {
        "name": "rescan_host",
        "help": "Rescan specific host ID",
        "func": _rescan_host,
        "args": [
            dict(arg_hostid),
        ],
    },
    {
        "name": "dt_stress",
        "help": "Run DT on specific path",
        "func": _dt_stress,
        "args": [
            dict(arg_of),
            dict(arg_runtime),
            dict(arg_log),
            dict(arg_threads),
        ],
    },
    {
        "name": "dt_stress_background",
        "help": "Run DT on background on specific path",
        "func": _dt_stress_background,
        "args": [
            dict(arg_of),
            dict(arg_runtime),
            dict(arg_log),
            dict(arg_threads),
        ],
    },
    {
        "name": "fio_stress",
        "help": "Run FIO on specific path",
        "func": _fio_stress,
        "args": [
            dict(arg_of),
            dict(arg_runtime),
            dict(arg_fio_verify),
            dict(arg_threads),
        ],
    },
    {
        "name": "fio_stress_background",
        "help": "Run FIO on background on specific path",
        "func": _fio_stress_background,
        "args": [
            dict(arg_of),
            dict(arg_runtime),
            dict(arg_fio_verify),
            dict(arg_threads),
        ],
    },
)


def main():
    parser = argparse.ArgumentParser(description="LibSAN tool")

    subparsers = parser.add_subparsers(help="Valid commands", dest="command")
    for cmd in cmds:
        _help = None
        if "help" in cmd:
            _help = cmd["help"]
        sub_parser = subparsers.add_parser(cmd["name"], help=_help)
        req_group = sub_parser.add_argument_group("required arguments")
        if "args" in cmd:
            for arg in cmd.get("args", []):
                name = arg.pop("name")
                required = False
                if "required" in arg:
                    required = arg.pop("required")
                if "dest" not in arg:
                    print("FAIL: dest needs to be set for arg: {} for command: {}".format(name, cmd["name"]))
                    sys.exit(1)
                name_list = name.split(" ")
                if required:
                    req_group.add_argument(*name_list, required=required, **arg)
                else:
                    sub_parser.add_argument(*name_list, required=required, **arg)

    args = parser.parse_args()

    for cmd in cmds:
        if cmd["name"] == args.command:
            if "func" not in cmd:
                print("FAIL: There is no function associated with %s" % cmd["name"])
                sys.exit(1)
            func = cmd["func"]
            func_args = vars(args)
            del func_args["command"]
            if func_args:
                if func(**func_args):
                    sys.exit(0)
            else:
                if func():
                    sys.exit(0)
            sys.exit(1)

    print("FAIL: Unsupported command: %s" % args.command)
    sys.exit(1)


main()



import nox


@nox.session(python=["3.6", "3.9", "3.11"])
def tests(session: nox.Session) -> None:
    """Runs pytest"""
    session.install("-e", ".[ssh]")
    session.install("pytest", "pytest-cov")
    session.run(
        "pytest",
        "--cov=libsan",
        "--cov-config",
        "pyproject.toml",
        "--cov-report=",
        *session.posargs,
        env={"COVERAGE_FILE": f".coverage.{session.python}"},
    )
    session.notify("coverage")


@nox.session
def coverage(session) -> None:
    """Coverage analysis"""
    session.install("coverage[toml]")
    session.run("coverage", "combine")
    session.run("coverage", "report")
    session.run("coverage", "erase")



# Copyright (C) 2018 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""stratis.py: Module to manipulate stratis userspace package."""

__author__ = "Jakub Krysl"
__copyright__ = "Copyright (c) 2018 Red Hat, Inc. All rights reserved."

import json

import libsan.host.linux
from libsan import _print
from libsan.host.cli_tools import Wrapper
from libsan.host.cmdline import run


def get_stratis_service() -> str:
    """Return name of the stratis service.
    :return: Name of the stratis service.
    :rtype: str
    """
    return "stratisd"


class Stratis(Wrapper):
    """Wrapper class for stratis command line interface."""

    def __init__(self, disable_check=True):
        self.stratis_version = ""
        self.disable_check = disable_check
        if libsan.host.linux.dist_ver() < 8:
            _print("FATAL: Stratis is not supported on RHEL < 8.")

        for pkg in ["stratisd", "stratis-cli"]:
            if not libsan.host.linux.is_installed(pkg) and not libsan.host.linux.install_package(pkg, check=False):
                _print(f"FATAL: Could not install {pkg} package")

        self.commands = {}
        self.commands["all"] = list(self.commands.keys())
        self.arguments = {
            "force": [self.commands["all"], " --force"],
            "redundancy": [self.commands["all"], " --redundancy"],
            "propagate": [self.commands["all"], "--propagate "],
        }

        if libsan.host.linux.service_status(get_stratis_service(), verbose=False) != 0:
            if not libsan.host.linux.service_restart(get_stratis_service()):
                _print(f"FAIL: Could not start {get_stratis_service()} service")
            else:
                _print(f"INFO: Service {get_stratis_service()} restarted.")

        Wrapper.__init__(self, self.commands, self.arguments, self.disable_check)

    @staticmethod
    def _remove_nones(kwargs):
        return {k: v for k, v in kwargs.items() if v is not None}

    def _run(self, cmd, verbosity=True, return_output=False, **kwargs):
        # Constructs the command to run and runs it

        # add '--propagate' flag by default to show whole trace when getting errors
        # This is a suggestion from devs
        if not ("propagate" in kwargs and not kwargs["propagate"]):
            cmd = self.arguments["propagate"][1] + cmd

        cmd = "stratis " + cmd
        cmd = self._add_arguments(cmd, **kwargs)

        ret = run(cmd, verbose=verbosity, return_output=return_output)
        if isinstance(ret, tuple) and ret[0] != 0:
            _print(f"WARN: Running command: '{cmd}' failed. Return with output.")
        elif isinstance(ret, int) and ret != 0:
            _print(f"WARN: Running command: '{cmd}' failed.")
        return ret

    def set_stratis_version(self, version):
        self.stratis_version = version

    def get_stratis_version(self):
        if not self.stratis_version:
            ret, data = self._run(cmd="--version", return_output=True)
            if ret == 0:
                self.set_stratis_version(data)
                return data
            print("FAIL: Could not get stratis version!")
            return "0.0.0"
        return self.stratis_version

    def get_stratis_major_version(self):
        return int(self.get_stratis_version().split(".")[0])

    def get_stratis_minor_version(self):
        return int(self.get_stratis_version().split(".")[1])

    def get_pool_uuid(self, pool_name):
        ret, data = self._run(cmd="report", return_output=True)
        if ret == 0 and data:
            try:
                report = json.loads(data)
            except json.JSONDecodeError:
                print("FAIL: Could not deserialize data returned from 'stratis report' command!")
                return None
            for pool in report["pools"]:
                if pool_name == pool["name"]:
                    print(f"INFO: Found UUID: {pool['uuid']} for pool_name: {pool_name}.")
                    return pool["uuid"]
            print(f"FAIL: Could not find pool UUID for provided pool_name: {pool_name}!")
            return None
        return None

    def get_fs_uuid(self, pool_name, fs_name):
        ret, data = self._run(cmd="report", return_output=True)
        if ret == 0 and data:
            try:
                report = json.loads(data)
            except json.JSONDecodeError:
                print("FAIL: Could not deserialize data returned from 'stratis report' command!")
                return ""
            for pool in report["pools"]:
                if pool_name != pool["name"]:
                    continue
                for fs in pool["filesystems"]:
                    if fs_name != fs["name"]:
                        continue
                    return fs["uuid"]
            print(f"FAIL: Could not find fs UUID for fs_name:{fs_name} from pool: {pool_name}!")
            return ""
        print("FAIL: Could not get stratis report.")
        return ""

    def pool_create(
        self,
        pool_name=None,
        blockdevs=None,
        force=False,
        redundancy=None,
        key_desc=None,
        tang_url=None,
        thumbprint=None,
        clevis=None,
        trust_url=None,
        no_overprovision=None,
        **kwargs,
    ):
        cmd = "pool create "
        if key_desc:
            cmd += f"--key-desc {key_desc} "
        if clevis:
            cmd += f"--clevis {clevis} "
        if tang_url:
            cmd += f"--tang-url {tang_url} "
        if thumbprint:
            cmd += f"--thumbprint {thumbprint} "
        if trust_url:
            cmd += "--trust-url "
        if no_overprovision:
            cmd += "--no-overprovision "
        if pool_name:
            cmd += f"{pool_name} "
        if blockdevs:
            if not isinstance(blockdevs, list):
                blockdevs = [blockdevs]
            cmd += " ".join(blockdevs)
        kwargs.update(
            {
                "force": force,
                "redundancy": redundancy,
            }
        )
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_list(self, pool_uuid=None, stopped_pools=None, **kwargs):
        cmd = "pool list "
        if pool_uuid:
            cmd += f"--uuid {pool_uuid} "
        if stopped_pools:
            cmd += "--stopped "
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_destroy(self, pool_name=None, **kwargs):
        cmd = "pool destroy "
        if pool_name:
            cmd += f"{pool_name} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_rename(self, current=None, new=None, **kwargs):
        cmd = "pool rename "
        if current:
            cmd += f"{current} "
        if new:
            cmd += f"{new} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_add_data(self, pool_name=None, blockdevs=None, **kwargs):
        cmd = "pool add-data "
        if pool_name:
            cmd += f"{pool_name} "
        if blockdevs:
            if not isinstance(blockdevs, list):
                blockdevs = [blockdevs]
            cmd += " ".join(blockdevs)
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_add_cache(self, pool_name=None, blockdevs=None, **kwargs):
        cmd = "pool add-cache "
        if pool_name:
            cmd += f"{pool_name} "
        if blockdevs:
            if not isinstance(blockdevs, list):
                blockdevs = [blockdevs]
            cmd += " ".join(blockdevs)
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_init_cache(self, pool_name=None, blockdevs=None, **kwargs):
        cmd = "pool init-cache "
        if pool_name:
            cmd += f"{pool_name} "
        if blockdevs:
            if not isinstance(blockdevs, list):
                blockdevs = [blockdevs]
            cmd += " ".join(blockdevs)
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_bind(
        self,
        binding_method=None,
        pool_name=None,
        key_desc=None,
        trust_url=None,
        thumbprint=None,
        tang_url=None,
        force=None,
        redundancy=None,
        **kwargs,
    ):
        cmd = "pool bind "
        if binding_method:
            cmd += f"{binding_method} "
        if trust_url:
            cmd += "--trust-url "
        if thumbprint:
            cmd += f"--thumbprint {thumbprint} "
        if pool_name:
            cmd += f"{pool_name} "
        if key_desc:
            cmd += f"{key_desc} "
        if tang_url:
            cmd += f"{tang_url} "
        kwargs.update(
            {
                "force": force,
                "redundancy": redundancy,
            }
        )
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_rebind(self, binding_method=None, pool_name=None, key_desc=None, force=None, redundancy=None, **kwargs):
        cmd = "pool rebind "
        if binding_method:
            cmd += f"{binding_method} "
        if pool_name:
            cmd += f"{pool_name} "
        if key_desc:
            cmd += f"{key_desc} "
        kwargs.update(
            {
                "force": force,
                "redundancy": redundancy,
            }
        )
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_unbind(self, binding_method=None, pool_name=None, force=None, redundancy=None, **kwargs):
        cmd = "pool unbind "
        if binding_method:
            cmd += f"{binding_method} "
        if pool_name:
            cmd += f"{pool_name} "
        kwargs.update(
            {
                "force": force,
                "redundancy": redundancy,
            }
        )
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_set_fs_limit(self, pool_name=None, fs_amount=None, force=None, redundancy=None, **kwargs):
        cmd = "pool set-fs-limit "
        if pool_name:
            cmd += f"{pool_name} "
        if fs_amount:
            cmd += f"{fs_amount} "
        kwargs.update(
            {
                "force": force,
                "redundancy": redundancy,
            }
        )
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_overprovision(self, pool_name=None, pool_overprovision=None, **kwargs):
        cmd = "pool overprovision "
        if pool_name:
            cmd += f"{pool_name} "
        if pool_overprovision:
            cmd += f"{pool_overprovision}"
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_explain(self, pool_error_code=None, **kwargs):
        cmd = "pool explain "
        if pool_error_code:
            cmd += f"{pool_error_code} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_debug(self, debug_subcommand=None, pool_name=None, pool_uuid=None, **kwargs):
        cmd = "pool debug "
        if debug_subcommand:
            cmd += f"{debug_subcommand} "
        if pool_name:
            cmd += f"--name {pool_name} "
        if pool_uuid:
            cmd += f"--uuid {pool_uuid} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_start(self, pool_name=None, pool_uuid=None, unlock_method=None, **kwargs):
        cmd = "pool start "
        if unlock_method:
            cmd += f"--unlock-method {unlock_method} "
        if self.get_stratis_major_version() <= 3 and self.get_stratis_minor_version() < 4:
            if pool_uuid:
                cmd += f"{pool_uuid} "
            return self._run(cmd, **self._remove_nones(kwargs))
        if pool_name:
            cmd += f"--name {pool_name} "
        if pool_uuid:
            cmd += f"--uuid {pool_uuid} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_stop(self, pool_name=None, **kwargs):
        cmd = "pool stop "
        if pool_name:
            cmd += f"{pool_name} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def pool_extend_data(self, pool_name=None, device_uuid=None, **kwargs):
        cmd = "pool extend-data "
        if pool_name:
            cmd += f"{pool_name} "
        if device_uuid:
            cmd += f"--device-uuid {device_uuid}"
        return self._run(cmd, **self._remove_nones(kwargs))

    def blockdev_list(self, pool_name=None, **kwargs):
        cmd = "blockdev list "
        if pool_name:
            cmd += f"{pool_name} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def blockdev_debug(self, debug_subcommand=None, dev_uuid=None, **kwargs):
        cmd = "blockdev debug "
        if debug_subcommand:
            cmd += f"{debug_subcommand} "
        if dev_uuid:
            cmd += f"--uuid {dev_uuid} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def fs_create(self, pool_name=None, fs_name=None, fs_size=None, **kwargs):
        cmd = "fs create "
        if fs_size:
            cmd += f"--size {fs_size} "
        if pool_name:
            cmd += f"{pool_name} "
        if fs_name:
            cmd += f"{fs_name} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def fs_debug(self, debug_subcommand=None, fs_name=None, fs_uuid=None, **kwargs):
        cmd = "fs debug "
        if debug_subcommand:
            cmd += f"{debug_subcommand} "
        if fs_name:
            cmd += f"--name {fs_name} "
        if fs_uuid:
            cmd += f"--uuid {fs_uuid} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def fs_snapshot(self, pool_name=None, origin_name=None, snapshot_name=None, **kwargs):
        cmd = "fs snapshot "
        if pool_name:
            cmd += f"{pool_name} "
        if origin_name:
            cmd += f"{origin_name} "
        if snapshot_name:
            cmd += f"{snapshot_name} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def fs_list(self, pool_name=None, **kwargs):
        cmd = "fs list "
        if pool_name:
            cmd += f"{pool_name} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def fs_destroy(self, pool_name=None, fs_name=None, **kwargs):
        cmd = "fs destroy "
        if pool_name:
            cmd += f"{pool_name} "
        if fs_name:
            cmd += f"{fs_name} "
        return self._run(cmd, **self._remove_nones(kwargs))

    def fs_rename(self, pool_name=None, fs_name=None, new_name=None, **kwargs):
        cmd = "fs rename "
        if pool_name:
            cmd += f"{pool_name} "
        if fs_name:
            cmd += f"{fs_name} "
        if new_name:
            cmd += f"{new_name} "
        return self._run(cmd, **kwargs)

    def key_set(self, keyfile_path=None, key_desc=None, **kwargs):
        cmd = "key set "
        if keyfile_path:
            cmd += f"--keyfile-path {keyfile_path} "
        if key_desc:
            cmd += f"{key_desc} "
        ret = self._run("key list", return_output=True)
        # ret is a tuple in format -> (return_code, return_output)
        if key_desc in ret[1]:
            return self._run("key list")
        return self._run(cmd, **kwargs)

    def key_reset(self, keyfile_path=None, key_desc=None, **kwargs):
        cmd = "key reset "
        if keyfile_path:
            cmd += f"--keyfile-path {keyfile_path} "
        if key_desc:
            cmd += f"{key_desc} "
        return self._run(cmd, **kwargs)

    def key_list(self, **kwargs):
        return self._run("key list", **kwargs)

    def key_unset(self, key_desc=None, **kwargs):
        cmd = "key unset "
        if key_desc:
            cmd += f"{key_desc} "
        return self._run(cmd, **kwargs)

    # Pool unlock has been removed in favor of pool_start in Stratisd 3.2.0
    def pool_unlock(self, **kwargs):
        return self._run("pool unlock", **kwargs)

    def daemon_redundancy(self, **kwargs):
        return self._run("daemon redundancy", **kwargs)

    def daemon_version(self, **kwargs):
        return self._run("daemon version", **kwargs)

    def debug_refresh(self, **kwargs):
        return self._run("debug refresh", **kwargs)

    def version(self, **kwargs):
        return self._run("--version", **kwargs)



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


"""fio.py: Module to run FIO util."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import os
import subprocess
import sys

import libsan.host.linux
from libsan import _print
from libsan.host.cmdline import exists, run

fio_default_options = {
    "rw": "randrw",  # Type of I/O pattern. Supported(read, write, trim, randread, randwrite, rw, randrw, trimwrite)
    "name": "fio_test",  # signalling the start of a new job.
    "filename": None,  # device or filename
    "direct": 1,  # If true, use non-buffered I/O (usually O_DIRECT)
    "iodepth": 1,  # Number  of  I/O  units  to keep in flight against the file. Note that increasing
    # iodepth beyond 1 will not affect synchronous ioengines
    "runtime": None,  # Terminate processing after the specified number of seconds.
    "size": None,  #
    "time_based": None,  # If given, run for the specified runtime duration even if the files are
    # completely read or written.
    "numjobs": 1,  # Number of clones (processes/threads performing the same workload) of this job
    "bs": "4k",  # lock  size for I/O units.
    "verify": None,  # Method  of verifying file contents after each iteration of the job
}  # (supports: md5 crc16 crc32 crc32c crc32c-intel crc64 crc7 sha256 sha512 sha1 xxhash)

fio_default_verify_options = {
    "verify_backlog": 1024,  # fio will write only N blocks before verifying these blocks.
    # Set to None to verify after all IO is written
    "verify_fatal": 1,  # If true, exit the job on the first observed verification failure
    "do_verify": 1,
}


def install_fio():
    pkg = "fio"
    if libsan.host.linux.install_package(pkg):
        return True
    if run("fio >/dev/null 2>&1") == 1:
        return True
    # Try to install FIO from source
    return install_fio_from_src()


def install_fio_from_src():
    git_url = "git://git.kernel.org/pub/scm/linux/kernel/git/axboe/fio.git"

    if not libsan.host.linux.install_package("libaio-devel"):
        _print("FAIL: Could not install libaio-devel")
        return False

    if not libsan.host.linux.install_package("zlib-devel"):
        _print("FAIL: Could not install zlib-devel")
        return False

    if run("git clone %s" % git_url) != 0:
        _print("FAIL: Could not clone fio repo")
        return False

    _print("INFO: Installing FIO")
    if run("cd fio && ./configure && make && make install") != 0:
        _print("FAIL: Could not build fio")
        return False

    if not exists("fio"):
        _print("FAIL: FIO did not install properly")
        return False
    return True


def fio_stress(of, verbose=False, return_output=False, **fio_opts):
    # For compatibilty with tests using other named parameters
    convert_param = {"io_type": "rw", "time": "runtime", "thread": "numjobs", "log_file": "output"}

    for key in convert_param:
        if key in list(fio_opts.keys()):
            fio_opts[convert_param[key]] = fio_opts.pop(key)

    fio_opts["filename"] = of

    for opt in fio_default_options:
        if opt not in list(fio_opts.keys()):
            fio_opts[opt] = fio_default_options[opt]

    if int(fio_opts["numjobs"]) > 1:
        fio_opts["group_reporting"] = 1

    if fio_opts["verify"] is not None:
        fio_opts.update(fio_default_verify_options)

    if not exists("fio"):
        _print("FATAL: fio is not installed")
        return False

    fio_param = ""
    for key in list(fio_opts.keys()):
        if fio_opts[key]:
            fio_param += f"--{key}='{fio_opts[key]}' "

    if "fiojob" in list(fio_opts.keys()):
        fio_param = "{} --filename={}".format(fio_opts["fiojob"], fio_opts["filename"])

    cmd = "fio %s" % fio_param
    # Append time information to command
    date = 'date "+%Y-%m-%d %H:%M:%S"'
    p = subprocess.Popen(date, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, _ = p.communicate()
    stdout = stdout.decode("ascii", "ignore")
    stdout = stdout.rstrip("\n")
    if not verbose:  # If verbose option is selected, the run() will print the fio command.
        _print(f"INFO: [{stdout}] FIO Running: '{cmd}'...")

    # _print("INFO: Running %s" % cmd)
    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("FAIL: running FIO")
        print(output)
        if return_output:
            return False, None
        return False

    _print("INFO: FIO executed successfully")
    if return_output:
        return True, output

    return True


def fio_stress_background(of, verbose=False, **fio_opts):
    """Run FIO on background"""
    newpid = os.fork()
    if newpid == 0:
        # Trying to flush stdout to avoid duplicated lines when running hba_test
        sys.stdout.flush()
        rt = fio_stress(of, verbose=verbose, **fio_opts)
        if not rt:
            os._exit(1)
        os._exit(0)
        return None
    else:
        sys.stdout.flush()
        _print("INFO: fio_stress_background(): Child thread %d is running FIO Stress" % newpid)
        return newpid



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""apconcli.py: Module to handle commands on ApCon Switch."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex
import socket
import sys
import time

import libsan.host.linux
import libsan.host.ssh
from libsan import _print


class apconcli:  # noqaN801
    """Class to manage ApCon switches"""

    host = None
    user = None
    passwd = None

    port_info_dict = None

    def __init__(self, hostname, username, password):
        self.host = hostname
        self.user = username
        self.passwd = password

    def _run(self, cmd, return_output=False, verbose=True, cr="\r", expect=">> ", command_expect=None):
        session = libsan.host.ssh.connect(self.host, user=self.user, passwd=self.passwd)
        if not session:
            _print("FAIL: Could not connect to switch (%s)" % self.host)
            if return_output:
                return 127, None
            return 127
        try:
            ret = libsan.host.ssh.run_cmd(
                session,
                cmd,
                return_output=return_output,
                verbose=verbose,
                invoke_shell=True,
                cr=cr,
                expect=expect,
                command_expect=command_expect,
            )
        except socket.timeout:
            _print("FAIL: Timeout connecting to %s" % self.host)
            if return_output:
                return 127, None
            return 127
        except Exception as e:
            _print("FAIL: Exception when connecting to switch")
            print("Exception is ", sys.exc_info()[0] + " " + e)
            if return_output:
                return 127, None
            return 127
        libsan.host.ssh.disconnect(session)
        return ret

    def get_version(self):
        """Return FW version of the switch"""
        cmd = "show version raw"
        _, output = self._run(cmd, return_output=True, verbose=False)
        if not output:
            _print("FAIL: get_version() - Could not get any output from switch")
            return None
        version_regex = re.compile(r".*Primary Cntrl F/W ver:[ \t]+([0-9]+)(.+)", re.MULTILINE)
        v = version_regex.search(output)
        if v:
            return "%s" % (v.group(1))
        _print("FAIL: get_version() - Could not parse output from switch")
        print(output)
        return None

    @staticmethod
    def capability():
        """Indicates supported operations on switch"""
        cap_dict = {}
        cap_dict["phy_port_connect"] = True
        cap_dict["phy_port_disconnect"] = True
        cap_dict["phy_port_flap"] = True
        cap_dict["phy_port_status"] = True
        cap_dict["phy_port_oscillate"] = True
        cap_dict["fc_physwitch"] = True
        return cap_dict

    def connect_state_dict(self):
        """Usage
            obj->connect_state_dict()
        Purpose
            Show the connection status and reture a dict like this:
                {A05: (06)}
            We depend on this command output:
                show connections raw
            Port 'A00' mean no connection defined, we omit them.
            We will use this info to reconnect ports if user not define all info.
        Parameter
            N/A
        Returns
            A dictionary of port ID to a list of Dest ports
        """
        cmd = "show connections raw"
        _, output = self._run(cmd, return_output=True, verbose=False)
        if not output:
            _print("FAIL: connect_state_dict() - Could not get any output from switch")
            return None

        state_dict = {}

        lines = output.split("\n")
        state_regex = re.compile(r"([A-Z]+[0-9]+):\s+([A-Z]+[0-9]+)")
        for line in lines:
            if re.search(cmd, line):
                # the output contains the command line, so we skip it
                continue
            m = state_regex.search(line)
            if m:
                src_port = m.group(1)
                dst_port = m.group(2)
                if dst_port == "A00":
                    # Skip A00 as it means no connection
                    continue
                if src_port not in list(state_dict.keys()):
                    state_dict[src_port] = []
                state_dict[src_port].append(dst_port)

        return state_dict

    def connect_mode_of(self, port_id):
        """Usage
            obj->connect_mode_of(port_id)
        Purpose
            Find out the connect stats for port_id.
            Return 'simplex' or 'duplex' or 'multicast' or None
        Parameter
            port_id        # like 'A01'
        Returns
            connect_state  # 'simplex' or 'duplex' or 'multicast'
                or
            None

        """
        if not port_id:
            _print("FAIL: connect_mode_of() - requires port_id as argument")
            return None

        port_state_dict = self.connect_state_dict()
        if port_id not in list(port_state_dict.keys()):
            return None

        n_conn = len(port_state_dict[port_id])
        for src_port in list(port_state_dict.keys()):
            for dst_port in port_state_dict[src_port]:
                if dst_port == port_id:
                    n_conn += 1

        if n_conn == 1:
            return "simplex"

        if n_conn == 2:
            return "duplex"

        if n_conn > 2:
            return "multicast"

        _print("FATAL: Unsupported mode on port %s" % port_id)
        return None

    def port_status(self, port_id):
        if not port_id:
            _print("FAIL: port_status() - needs 'port_id' parameter")
            return False

        cmd = "show port info %s" % port_id
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            # It will never hit this point as _run with invoke_shell always return success
            _print("FAIL: port_status() - command failed")
            return None

        status_regex = re.compile(r"%s:\s+Status:\s+(\S+)" % port_id)
        lines = output.split("\n")
        for line in lines:
            m = status_regex.match(line)
            if m:
                return m.group(1)
        _print("FAIL: port_status() - Could not find status for port %s" % port_id)
        print(output)
        return None

    def port_disconnect(self, port_id, check_state=True):
        if not port_id:
            _print("FAIL: port_disconnect() - needs 'port_id' parameter")
            return False

        cmd = "disconnect %s" % port_id
        ret = self._run(cmd, verbose=False)
        if ret != 0:
            # It will never hit this point as _run with invoke_shell always return success
            _print("FAIL: port_disconnect() - command failed")
            return False

        if not check_state:
            _print("INFO: Successfully disconnected port %s" % port_id)
            return True

        state_dict = self.connect_state_dict()
        if port_id not in list(state_dict.keys()):
            _print("INFO: Successfully disconnected port %s" % port_id)
            return True

        _print("FAIL: port_disconnect() - Failed to disconnect port %s." % port_id)
        return False

    def port_connect(self, src_port_id, dst_port_id, mode, check_state=True):
        if not src_port_id or not dst_port_id or not mode:
            _print("FAIL: port_connect() - needs 'src_port_id' , 'dst_port_id' and 'mode' parameters")
            return False

        if mode != "simplex" and mode != "duplex":
            _print("FAIL: mode %s is not supported" % mode)
            return False

        cmd = f"connect {mode} {src_port_id} {dst_port_id}"
        ret = self._run(cmd, verbose=False)
        if ret != 0:
            # It will never hit this point as _run with invoke_shell always return success
            _print("FAIL: port_connect() - command failed")
            return False

        if not check_state:
            _print(f"INFO: Successfully connected port {src_port_id} to {dst_port_id} ({mode})")
            return True

        connected_mode = self.connect_mode_of(src_port_id)
        if connected_mode and connected_mode == mode:
            _print(f"INFO: Successfully connected port {src_port_id} to {dst_port_id} ({mode})")
            return True

        _print(f"FAIL: port_connect() - Failed to connect port {src_port_id} to {dst_port_id} ({mode}).")
        return False

    def port_oscillate(self, src_port_id, dst_port_id, mode, min_uptime, max_uptime, min_downtime, max_downtime, count):
        if (
            not src_port_id
            or not dst_port_id
            or not mode
            or min_uptime is None
            or max_uptime is None
            or min_downtime is None
            or max_downtime is None
            or count is None
        ):
            _print(
                "FAIL: port_oscillate() - needs 'src_port_id' , 'dst_port_id', 'mode', 'min_uptime',"
                " 'max_uptime', 'min_downtime', 'max_downtime' and 'count' parameters"
            )
            return False

        for cnt in range(1, count + 1):
            _print("INFO: Interaction %d/%d" % (cnt, count))
            self.port_disconnect(src_port_id, check_state=False)
            increment = (max_downtime - min_downtime) / count
            sleep_time = min_downtime + ((cnt - 1) * increment)
            _print("INFO: Going to sleep for %s with link down" % sleep_time)
            libsan.host.linux.sleep(sleep_time)
            self.port_connect(src_port_id, dst_port_id, mode, check_state=False)
            increment = (max_uptime - min_uptime) / count
            sleep_time = min_uptime + ((cnt - 1) * increment)
            _print("INFO: Going to sleep for %s with link up" % sleep_time)
            libsan.host.linux.sleep(sleep_time)

        return True

    def port_flap(self, port_id, uptime, downtime, count):
        """Purpose
            Flap a port via command 'flap'.
            We will sleep utils flap done. 'flap' command in ApConCLI is still
            interactive command.
            The maximum duration of a flap is 10 seconds.
            So if that exceed, we will reduce count first.
            If uptime + downtime exceeded 10 seconds, we will change each of them to 5s.
        Parameter
            port_id        # like 'A05'
            uptime         # time to keep link up during flap in micro seconds
            downtime       # time to keep link down during flap in micro seconds
            count          # how many times should we flap the port. like '10'
        """
        if not port_id or not uptime or not downtime or not count:
            _print("FAIL: port_flap() - requires 'port_id', 'uptime', 'downtime' and count as parameters")
            return False

        # make sure uptime, downtime, count are int
        try:
            uptime = int(uptime)
            downtime = int(downtime)
            count = int(count)
        except Exception as e:
            _print("FAIL: uptime, downtime and count must be integers")
            print(e)
            return False

        if uptime + downtime > 10000:
            _print("WARN: port_flap() - The sum of uptime and downtime exceed 10s, chaging each of them to 5s")
            uptime = 5000
            downtime = 5000

        # Might be better to use expect
        cmds = ["flap", "%s" % port_id, "%s" % downtime, "%s" % uptime, "%s" % count]
        ret, output = self._run(cmds, return_output=True, verbose=False, command_expect=": ")
        if ret != 0:
            # It will never hit this point as _run with invoke_shell always return success
            _print("FAIL: port_flap() - command failed")
            return False

        flap_start_regex = re.compile("Port flapping started", re.MULTILINE)
        if output and flap_start_regex.search(output):
            sleep_time = int((uptime + downtime) * count / 1000) + 1
            _print(
                "INFO: port_flap() - Flapping port %s uptime %s ms, downtime %s ms, count %s"
                % (port_id, uptime, downtime, count)
            )
            _print(f"INFO: port_flap() - Waiting {sleep_time} seconds while flapping port {port_id}")
            time.sleep(sleep_time)
            _print("INFO: port_flap() - Port %s finished flapping" % port_id)
            return True

        _print("FAIL: Could not start flap on port %s" % port_id)
        print(output)
        return False



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""loopdev.py: Module to manipulate loop devices using losetup."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import os
import sys

import libsan.host.linux
import libsan.misc.size
from libsan import _print
from libsan.host.cmdline import run


def _get_loop_path(name):
    loop_path = name
    if "/dev/" not in name:
        loop_path = "/dev/" + name

    return loop_path


def _get_image_file(name, image_path):
    image_file = f"{image_path}/{name}.img"
    return image_file


def _standardize_name(name):
    """Make sure use same standard for name, for example remove /dev/ from it if exists"""
    if not name:
        _print("FAIL: _standardize_name() - requires name as parameter")
        return None
    return name.replace("/dev/", "")


def create_loopdev(name=None, size=1024, image_path="/var/tmp", reuse_file=False):
    """Create a loop device
    Parameters:
    \tname:         eg. loop0 (optional)
    \tsize:         Size in MB (default: 1024MB)
    \timage_path:   Path to store the image (default: /var/tmp)
    \treuse_file:   Reuse a previous image file (default: False)
    """
    #    name = args['name']
    #    if name is None:
    #        name = "loop0"

    #    size = args['size']
    #    if size is None:
    #        size = 1024

    if not name:
        cmd = "losetup -f"
        retcode, output = run(cmd, return_output=True, verbose=False)
        if retcode != 0:
            _print("FAIL: Could not find free loop device")
            print(output)
            return None
        name = output
    name = _standardize_name(name)

    fname = _get_image_file(name, image_path)
    print("INFO: Creating loop device %s with size %d" % (fname, size))

    if not reuse_file and os.path.isfile(fname):
        print("INFO: Deleting file %s" % fname)
        # If for some reason the file exist and we don't want reuse, delete it.
        run("rm -f %s" % fname)

    # make sure we have enough space to create the file
    free_space_bytes = libsan.host.linux.get_free_space(image_path)
    # Convert the size given in megabytes to bytes
    size_bytes = int(libsan.misc.size.size_human_2_size_bytes("%sMiB" % size))
    if free_space_bytes <= size_bytes:
        _print(
            "FAIL: Not enough space to create loop device with size %s"
            % libsan.misc.size.size_bytes_2_size_human(size_bytes)
        )
        _print("available space: %s" % libsan.misc.size.size_bytes_2_size_human(free_space_bytes))
        return None
    print("INFO: Creating file %s" % fname)
    # cmd = "dd if=/dev/zero of=%s seek=%d bs=1M count=0" % (fname, size)
    cmd = f"fallocate -l {size}M {fname}"
    try:
        # We are just creating the file, not writting zeros to it
        retcode = run(cmd)
        if retcode != 0:
            _print("command failed with code %s" % retcode)
            _print("FAIL: Could not create loop device image file")
            return None
    except OSError as e:
        print("command failed: ", e, file=sys.stderr)
        return None

    loop_path = _get_loop_path(name)
    # detach loop device if it exists
    detach_loopdev(loop_path)

    # Going to associate the file to the loopdevice
    cmd = f"losetup {loop_path} {fname}"
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: Could not create loop device")
        return None

    return loop_path


def delete_loopdev(name):
    """Delete a loop device
    Parameters:
    \tname:     eg. loop0 or /dev/loop0
    """
    if not name:
        _print("FAIL: delete_loopdev() - requires name parameter")
        return False

    print("INFO: Deleting loop device %s" % name)
    name = _standardize_name(name)

    loop_path = _get_loop_path(name)

    # find image file
    fname = get_loopdev_file(loop_path)
    if fname is None:
        _print("WARN: could not find loopdev named %s" % name)
        # loopdev does not exist, nothing to do
        return True

    # detach loop device if it exists
    if not detach_loopdev(name):
        _print("FAIL: could not detach %s" % loop_path)
        return False

    if os.path.isfile(fname):
        cmd = "rm -f %s" % fname
        retcode = run(cmd)
        if retcode != 0:
            _print("FAIL: Could not delete loop device file %s" % fname)
            return False

    # check if loopdev file is deleted as it sometimes remains
    if os.path.isfile(fname):
        _print("FAIL: Deleted loop device file %s but it is still there" % fname)
        return False

    return True


# show loop devices
def list_loopdev():
    retcode, output = run("losetup -a", return_output=True)
    return retcode, output


# Return all loop devices
def get_loopdev():
    # example of output on rhel-6.7
    # /dev/loop0: [fd00]:396428 (/var/tmp/loop0.img)
    retcode, output = run("losetup -a | awk '{print$1}'", return_output=True, verbose=False)
    # retcode, output = run("losetup -l | tail -n +2", return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: get_loopdev failed to execute")
        print(output)
        return None

    devs = None
    if output:
        devs = output.split("\n")
        # remove the ":" character from all devices
        devs = [d.replace(":", "") for d in devs]

    return devs


# Return loop device file for given path
def get_loopdev_file(loop_path):
    # example of output on rhel-6.7
    # /dev/loop0: [fd00]:396428 (/var/tmp/loop0.img)
    retcode, output = run("losetup -a | grep '%s:' | awk '{print$3}'" % loop_path, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: get_loopdev_file failed to execute")
        print(output)
        return None

    if output:
        # remove the "(" and ")" character from device
        dev = output[1:-1]
    else:
        _print("WARN: get_loopdev_file failed to requested loopdev")
        return None

    return dev


def detach_loopdev(name=None):
    cmd = "losetup -D"
    if name:
        devs = get_loopdev()
        if not devs:
            # No device was found
            return False

        name = _standardize_name(name)

        # Just try to detach if device is connected, otherwise ignore
        # print "INFO: Checking if ", loop_path, " exists, to be detached"
        dev_path = _get_loop_path(name)
        if dev_path in devs:
            cmd = "losetup -d %s" % dev_path
        else:
            # if loop device does not exist just ignore it
            return True

    # run losetup -D or -d <device>
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: Could not detach loop device")
        return False

    return True



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""time.py: Module to convert time."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re
from datetime import datetime

from libsan import _print


def time_2_sec(time_str):
    """Usage
        time_2_sec(time_str)
    Purpose
        Convert time_str into seconds. Support days, hours, minutes and seconds
        in this format:
            1h2m3s
    Parameter
        time_str       # like "1d2h3m4s"
    Return
        time in seconds:    like 93784
    """
    if time_str is None:
        _print("FAIL: time_2_sec() - requires time_str parameter")
        return None

    time_str = str(time_str)
    if re.match(r"\d+$", time_str):
        return int(time_str)

    seconds = None

    # for example 4d
    m = re.search(r"(\d+)d", time_str)
    if m:
        if seconds is None:
            seconds = 0
        seconds += 24 * 60 * 60 * int(m.group(1))

    # for example 2h
    m = re.search(r"(\d+)h", time_str)
    if m:
        if seconds is None:
            seconds = 0
        seconds += 60 * 60 * int(m.group(1))

    # for example 1m
    m = re.search(r"(\d+)m", time_str)
    if m:
        if seconds is None:
            seconds = 0
        seconds += 60 * int(m.group(1))

    # for example 10s
    m = re.search(r"(\d+)s", time_str)
    if m:
        if seconds is None:
            seconds = 0
        seconds += int(m.group(1))

    if seconds is None:
        _print("FAIL: time_2_sec() - Could not parse '%s'" % time_str)
        return None

    return seconds


def sec_2_time(seconds):
    """Usage
        sec_2_time(seconds)
    Purpose
        Convert time in seconds to a more human format.
    Parameter
        seconds       # like 93784
    Return
        time in human format:    like "1d2h3m4s"
    """
    if seconds is None:
        _print("FAIL: sec_2_time() - requires seconds as parameter")
        return None

    # it can't be converted to int
    if not re.match(r"\d+$", str(seconds)):
        _print("FAIL: sec_2_time() - %s is not a valid parameter" % seconds)
        return None

    # make sure it is int
    seconds = int(seconds)

    secs = seconds % 60
    time_str = "%02ds" % secs
    if seconds > 59:
        minutes = seconds / 60
        mins = minutes % 60
        time_str = "%02dm%s" % (mins, time_str)
        if minutes > 59:
            hours = minutes / 60
            hrs = hours % 24
            time_str = "%02dh%s" % (hrs, time_str)
            if hours > 23:
                days = hours / 24
                time_str = "%02dd%s" % (days, time_str)
    return time_str


def get_time(in_seconds=False):
    """Get the current time
    Parameter
        in_seconds       # Boolean: Return time in second (default: False)
    Return
        Current time:   #By default as 20160714034418 (yyyymmddHHMMSS)
    """
    now = datetime.now()

    ts = now.strftime("%Y%m%d%H%M%S")
    if in_seconds:
        ts = int(now.strftime("%s"))
    return ts



#!/usr/bin/python

# Copyright (C) 2021 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


import unittest
from unittest.mock import patch

from libsan.host import lvm


class TestPV(unittest.TestCase):
    pvs_output = "  /dev/vda1,lvm_test,lvm2,a--,98.41g,0"
    pv_query_output = {"/dev/vda1": {"vg": "lvm_test", "fmt": "lvm2", "attr": "a--", "psize": "98.41g", "pfree": "0"}}

    def test_pv_query(self):
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = [0, self.pvs_output]
            assert self.pv_query_output == lvm.pv_query()

    def test_pv_create(self):
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = 0
            assert lvm.pv_create("/dev/vda1")
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = 1
            assert not lvm.pv_create("/dev/vda1")

    @patch("libsan.host.lvm.pv_query")
    def test_pv_remove(self, mock):
        mock.return_value = self.pv_query_output
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = 0
            assert lvm.pv_remove("/dev/vda1")
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = 1
            assert not lvm.pv_remove("/dev/vda1")


class TestVG(unittest.TestCase):
    vgs_output = "  lvm_test,1,3,0,wz--n-,98.41g,0"
    vg_query_output = {
        "lvm_test": {"num_pvs": "1", "num_lvs": "3", "num_sn": "0", "attr": "wz--n-", "vsize": "98.41g", "vfree": "0"}
    }

    def test_vg_query(self):
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = [0, self.vgs_output]
            assert self.vg_query_output == lvm.vg_query()

    def test_pv_create(self):
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = 0
            assert lvm.vg_create("lvm_test", "/dev/vda1")
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = 1
            assert not lvm.vg_create("lvm_test", "/dev/vda1")

    @patch("libsan.host.lvm.pv_query")
    def test_pv_remove(self, mock):
        mock.return_value = self.vg_query_output
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = 0
            assert lvm.pv_remove("lvm_test")
        with patch("libsan.host.lvm.run") as run_func:
            run_func.return_value = 1
            assert not lvm.pv_remove("lvm_test")



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


"""eqlogic.py: Module to handle commands on EqualLogic Array."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex
import sys

import libsan.array.dell.eqlparse_show
import libsan.host.conf
import libsan.host.ssh
import libsan.misc.size
from libsan import _print


def _size_human_2_size_eqlogic(size):
    """Convert size to size Equallogic expects. Eg 2GiB to 2G"""
    size_human = libsan.misc.size.size_bytes_2_size_human(size)
    if not libsan.misc.size.size_human_check(size_human):
        _print("FAIL: %s is not on human size format")
        return None
    size_eql = re.sub("iB", "", size_human)
    return size_eql


def _size_eqlogic_2_size_human(size_eql):
    """Convert size Equallogic to Human. Eg 2GB to 2GiB"""
    if not size_eql:
        return None
    m = re.match(r"(\S+[M|G|T])B?$", size_eql)
    if not m:
        _print("FAIL: _size_eqlogic_2_size_human() - Invalid Eqlogic size format: %s" % size_eql)
        return None
    size_human = "%siB" % m.group(1)
    return size_human


def parse_ret(data):
    # remove command on 1st line and last line with 'grub>'
    return data.split("\r\n")[1:-1]


class eqlogic:  # noqa: N801
    """Class to manage Equallogic array"""

    host = None
    user = None
    passwd = None

    san_conf_path = None
    san_dev_name = None

    volume_dict = None
    member_dict = None
    luns: list = []

    map_info: list = []

    def __init__(self, hostname, username, password, timeout=None, san_dev=None):
        self.host = hostname
        self.user = username
        self.passwd = password
        self.timeout = timeout
        if self.timeout:
            self.timeout = float(self.timeout)
        self.san_conf_path = None
        # san_dev is the nake on san_top for this device
        self.san_dev_name = san_dev
        self.sa_conf_dict = None

    def set_san_conf_path(self, san_conf_path):
        self.san_conf_path = san_conf_path
        return True

    def set_sa_conf(self, sa_conf_dict):
        self.sa_conf_dict = sa_conf_dict
        return True

    def get_sa_conf(self):
        return self.sa_conf_dict

    def _run(self, cmd, return_output=False, verbose=True):
        # Create the session and log in
        session = libsan.host.ssh.connect(self.host, user=self.user, passwd=self.passwd)
        if not session:
            _print("FAIL: Could not connect to switch (%s)" % self.host)
            if return_output:
                return 127, None
            return 127

        try:
            ret = libsan.host.ssh.run_cmd(
                session,
                cmd=cmd,
                return_output=return_output,
                verbose=verbose,
                timeout=self.timeout,
                cr="\n\r",
                invoke_shell=True,
                expect="> ",
            )
        except Exception as e:
            print(e)
            print("Unexpected error:", sys.exc_info()[0])
            # _print("FAIL: Timeout executing command on address: %s" % (self.host))
            print(cmd)
            if return_output:
                return 1, "Exception on ssh"
            return 1
        libsan.host.ssh.disconnect(session)

        if return_output:
            ret_code, output = ret
            output = "\n".join(parse_ret(output))
            return ret_code, output

        return ret

    def query_members(self, recheck=False):
        """Query all LUNs on array and store its information on a dict"""

        # If we do not need to recheck and port info exist we return it
        if self.member_dict and not recheck:
            return self.member_dict

        cmd = "member show"
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: "%s" command' % cmd)
            if output:
                print(output)
            return None
        self.member_dict = libsan.array.dell.eqlparse_show.parse_show(output)

        return self.member_dict

    def get_version(self):
        """Return FW version of the array"""
        member_dict = self.query_members()
        if not member_dict:
            return None

        # Return the version of the first group we found
        for key in member_dict:
            if "version" in member_dict[key]:
                return member_dict[key]["version"]
            return None

    @staticmethod
    def capability():
        """Indicates supported operation on array"""
        cap_dict = {}
        cap_dict["lun_info"] = True
        cap_dict["lun_query"] = True
        cap_dict["lun_create"] = True
        cap_dict["lun_map"] = True
        cap_dict["lun_unmap"] = True
        cap_dict["lun_remove"] = True
        cap_dict["lun_grow"] = True
        cap_dict["lun_shrink"] = True
        cap_dict["lun_trepass"] = False
        cap_dict["lun_thinp"] = True
        cap_dict["sa_ctrler_reboot"] = False
        return cap_dict

    def query_all_luns(self):
        """Query all LUNs on array and store its information on a dict"""

        # If we do not need to recheck and port info exist we return it
        # if self.luns and not recheck:
        #    return self.luns

        cmd = "volume show -volume"
        _print(f"INFO: Running command {cmd} on {self.host}")
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: "%s" command' % cmd)
            print(output)
            return None
        _print("INFO: finished querying all volumes")
        output_dict = libsan.array.dell.eqlparse_show.parse_show(output)
        if not output_dict:
            return None

        self.luns = []
        for lun_name in output_dict:
            self.luns.append(lun_name)

        return sorted(self.luns)

    def query_all_lun_info(self):
        """Query all LUNs on array and store its information on a dict"""

        # If we do not need to recheck and port info exist we return it
        # if self.volume_dict and not recheck:
        #    return self.volume_dict

        luns = self.query_all_luns()
        if not luns:
            return None
        _print("INFO: Going to query info for each LUN, this might take a while...")
        for lun_name in luns:
            if not self.volume_dict:
                self.volume_dict = {}
            self.volume_dict[lun_name] = self.lun_info(lun_name)
            # lun_info = self.volume_dict[lun_name]
            # lun_info["name"] = lun_name
        _print("INFO: finished querying all volumes info")
        return self.volume_dict

    def _lun_exist(self, lun_name):
        """Test if LUN exist
        If LUN does not exist return error
        """
        if not lun_name:
            _print("FAIL: _lun_exist - requires lun_name")
            return False

        # cmd = "volume show %s" % lun_name
        # ret, output = self._run(cmd, return_output = True, verbose = False)
        # if ret != 0:
        #    return False
        # print output
        # return True
        if self.lun_info(lun_name):
            return True
        return False

    @staticmethod
    def _lun_wwid(lun_info_dict):
        """From the LUN iSCSI name it is possible to convert it to wwid"""
        if not lun_info_dict:
            return None
        # wwid = None
        lun_name = lun_info_dict["name"]
        # Last \S+ usually is the lun name, but could be different if volume was renamed...
        # Has to be non greedy match
        eql_wwid_iqn_regex = re.compile(r"^iqn\..*equallogic:(\S+?)-(\S+?)-(\S+?)-(\S+?)-\S+$")
        m = eql_wwid_iqn_regex.match(lun_info_dict["iscsi_name"])
        if m:
            match1 = m.group(1)
            match2 = m.group(2)
            match3 = m.group(3)
            match4 = m.group(4)
            if len(match4) % 2 != 0:
                match4 += match3[0]
                # as we added the character to the other match, we remove from this one
                match3 = match3[1:]
            if len(match3) % 2 != 0:
                match3 += match2[0]
                # as we added the character to the other match, we remove from this one
                match2 = match2[1:]
            if len(match2) % 2 != 0:
                match2 += match1[0]
                # as we added the character to the other match, we remove from this one
                match1 = match1[1:]
            if len(match1) % 2 != 0:
                match1 += "Z"
                print(lun_info_dict["iscsi_name"])
                _print("FATAL: _lun_wwid() could not parse wwid")
            match1 = "".join(reversed([match1[i : i + 2] for i in range(0, len(match1), 2)]))
            match2 = "".join(reversed([match2[i : i + 2] for i in range(0, len(match2), 2)]))
            match3 = "".join(reversed([match3[i : i + 2] for i in range(0, len(match3), 2)]))
            match4 = "".join(reversed([match4[i : i + 2] for i in range(0, len(match4), 2)]))
            wwid = "3" + match1 + match2 + match3 + match4
        else:
            print(lun_info_dict)
            _print("FAIL: _lun_wwid() - Could not get wwid of %s" % lun_name)
            return None
            # _print("DEBUG: lun_wwid: wwid %s" % wwid)
        return wwid

    def lun_info(self, lun_name):
        """Query detailed information of specific LUN"""
        # used to match regex for each session information that we support
        # supported_lun_info = {"name"    : "^Name: (\S+)",
        #                    "size_human": "^Size: (\S+)",
        #                    "iscsi_name": "^iSCSI Name: (\S+)",
        #                    "status"    : "^Status: (\S+)"}
        supported_lun_info = ["name", "size", "iscsi_name", "status"]

        if not lun_name:
            _print("FAIL: lun_info - requires lun_name")
            return None

        cmd = "volume show %s" % lun_name
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print("FAIL: lun_info() - lun does not exist")
            return None

        if not output:
            return None

        output_dict = libsan.array.dell.eqlparse_show.parse_show(output)
        if not output_dict:
            return None

        if "_volume_information_" not in output_dict:
            return None
        vol_info_dict = output_dict["_volume_information_"]
        lun_info_dict = {}
        for key in vol_info_dict:
            if key in supported_lun_info:
                lun_info_dict[key] = vol_info_dict[key]

        lun_info_dict["size_human"] = _size_eqlogic_2_size_human(lun_info_dict["size"])
        lun_info_dict["size_eqlogic"] = lun_info_dict["size"]
        # Replace eqlogic size for size in bytes
        lun_info_dict["size"] = libsan.misc.size.size_human_2_size_bytes(lun_info_dict["size_human"])

        if "_access_records_" in output_dict:
            map_infos = output_dict["_access_records_"]
            map_info = None
            for key in map_infos:
                map_info = {}
                map_info["t_iqn"] = lun_info_dict["iscsi_name"]
                map_info["h_iqn"] = map_infos[key]["initiator"]
                map_info["id"] = map_infos[key]["id"]
                if "map_infos" not in lun_info_dict:
                    lun_info_dict["map_infos"] = []
            lun_info_dict["map_infos"].append(map_info)
        # lines = output.split("\n")
        # lun_info_dict = None

        # line_id = 0
        # n_lines = len(lines)
        # while line_id < n_lines:
        # line = lines[line_id].rstrip('\n')
        # line_id += 1
        # #print "DEBUG parse: try %s - %s" % (line)
        # if not line:
        # continue

        # for key in supported_lun_info.keys():
        # m = re.match(supported_lun_info[key], line)
        # if not m:
        # continue
        # if not lun_info_dict:
        # lun_info_dict = {}
        # #print "Found %s: %s" % (key, m.group(1))
        # lun_info_dict[key] = m.group(1)
        # #If next line does not look like a parameter, it is probably because
        # #Equallogic wrapped the previous line, for example on iSCSI alias scenario
        # if (line_id) < n_lines:
        # next_line = lines[line_id].rstrip('\n')
        # if not re.search(": ", next_line):
        # lun_info_dict[key] += next_line

        if "iscsi_name" in lun_info_dict:
            wwid = self._lun_wwid(lun_info_dict)
            if wwid:
                lun_info_dict["wwid"] = wwid

        return lun_info_dict

    def lun_create(self, name, size, permission="read-write", status="online", sec_size="512"):
        """Create new LUN on array if LUN does not exist"""
        if not name or not size:
            _print("FAIL: usage lun_create(name, size)")
            return None

        size_eql = _size_human_2_size_eqlogic(size)
        if not size_eql:
            _print("FAIL: Could not get human size %s on Equallogic format" % size)
            return None

        if self._lun_exist(name):
            _print("FAIL: lun_create() - LUN %s already exist" % name)
            return None

        cmd = f"volume create {name} {size_eql} {permission} {status} sector-size {sec_size}"
        ret, output = self._run(cmd, return_output=True, verbose=True)
        if ret != 0:
            _print('FAIL: lun_create() - "%s" command' % cmd)
            print(output)
            return None
        # Need to update our volumes
        # luns = self.query_all_luns(recheck=True)
        # if name not in luns:
        if not self._lun_exist(name):
            _print("FAIL: lun_create() - thought LUN %s was created, but it was not" % name)
            return None

        # As new target is created, we need to update san conf file
        # _print("DEBUG: lun_create san_conf_file: %s" % self.san_conf_path)
        info_dict = self.lun_info(name)
        if not self.san_conf_path:
            _print("WARN: lun_create() - Could not find path for san_conf file")
            return name

        if not self.san_dev_name:
            _print(f"WARN: lun_create() - it seems {self.host} is not configured on {self.san_conf_path}")
        else:
            libsan.host.conf.config_add_entry(
                self.san_dev_name, "iscsi_iqn_%s" % name, info_dict["iscsi_name"], self.san_conf_path
            )
        return name

    def lun_remove(self, name):
        """Delete a LUN on array if LUN exists"""
        if not name:
            _print("FAIL: usage lun_remove(name)")
            return False

        if not self._lun_exist(name):
            _print("INFO: LUN %s do NOT exist" % name)
            return True

        vol_dict = self.lun_info(name)
        if not vol_dict:
            _print("FAIL: Could not get info for %s" % name)
            return False

        if vol_dict["status"] != "offline":
            _print("INFO: Setting LUN %s offline before deleting it..." % name)
            if not self._lun_change_status(name, "offline"):
                _print("FAIL: Could not set %s offline")
                return False

        cmd = "volume delete %s" % name
        ret, output = self._run(cmd, return_output=True, verbose=True)
        if ret != 0:
            _print('FAIL: lun_remove() - "%s" command' % cmd)
            print(output)
            return False
        # Need to update our volumes
        # luns = self.query_all_luns(recheck=True)
        # if name in luns:
        if self._lun_exist(name):
            _print("FAIL: lun_remove() - thought LUN %s was deleted, but it was not" % name)
            return False

        if not self.san_conf_path:
            _print("WARN: lun_remove() - Could not find path for san_conf file")
            return True

        if not self.san_dev_name:
            _print(f"WARN: lun_remove() - it seems {self.host} is not configured on {self.san_conf_path}")
        else:
            # As lun is remove so is the target, we need to update san conf file
            libsan.host.conf.config_remove_entry(self.san_dev_name, "iscsi_iqn_%s" % name, self.san_conf_path)
        return True

    def lun_map(self, name):
        """Map a LUN to an IQN"""
        if not name:
            _print("FAIL: lun_map() - requires lun_name as parameter")
            return False

        if not self.map_info:
            _print("FAIL: lun_map() - There is no mapping info to use")
            return False

        # vol_dict = self.query_all_lun_info()
        # if name not in vol_dict.keys():
        if not self._lun_exist(name):
            _print("FAIL: lun_map - LUN %s do NOT exist" % name)
            return False

        iqns_2_map = []
        for map_info in self.map_info:
            if "h_iqn" not in map_info.keys():
                _print("FAIL: lun_map() - Does not know how to map using")
                print(map_info)
                continue
            if map_info["h_iqn"] not in iqns_2_map:
                iqns_2_map.append(map_info["h_iqn"])

        map_infos = self.lun_query_access(name)
        for iqn in iqns_2_map:
            # Check if this IQN is already mapped to the LUN
            if map_infos:
                for map_info in map_infos:
                    if "h_iqn" not in map_info.keys():
                        continue
                    if iqn == map_info["h_iqn"]:
                        _print(f"WARN: LUN {name} is already mapped to {iqn}")
                        return True

            cmds = ["volume select %s" % name, "access create initiator %s" % iqn]
            ret, output = self._run(cmds, return_output=True, verbose=True)
            if ret != 0:
                _print('FAIL: "%s" command' % ", ".join(cmds))
                print(output)
                return False
        # Update map details
        # self.volume_dict[name] = self.lun_info(name)
        return True

    def lun_unmap(self, name):
        """Unmap a LUN to an IQN"""
        if not name:
            _print("FAIL: usage lun_unmap(name)")
            return False

        # vol_dict = self.query_all_lun_info()
        # if name not in vol_dict.keys():
        if not self._lun_exist(name):
            _print("FAIL: lun_unmap - LUN %s do NOT exist" % name)
            return False

        if not self.map_info:
            _print("FAIL: lun_map() - There is no mapping info to use")
            return False

        iqns_2_unmap = []
        for map_info in self.map_info:
            if "h_iqn" not in map_info.keys():
                _print("FAIL: lun_map() - Does not know how to map using")
                print(map_info)
                continue
            if map_info["h_iqn"] not in iqns_2_unmap:
                iqns_2_unmap.append(map_info["h_iqn"])

        # access_dict = self.lun_query_access(name)
        map_infos = self.lun_query_access(name)
        if not map_infos:
            _print("FAIL: Could not find any map information for this LUN")
            return False
        # To unmap we need to know the ID of the IQN
        access_id = None
        error = 0
        for iqn in iqns_2_unmap:
            # Check if this IQN is mapped to the LUN
            for map_info in map_infos:
                if "h_iqn" not in map_info.keys():
                    continue
                if iqn == map_info["h_iqn"]:
                    access_id = map_info["id"]

                if not access_id:
                    _print(f"WARN: LUN {name} is not mapped to {iqn}")
                    return True

                cmds = ["volume select %s" % name, "access delete %s" % access_id]
                ret, output = self._run(cmds, return_output=True, verbose=True)
                if ret != 0:
                    _print('FAIL: "%s" command' % ", ".join(cmds))
                    print(output)
                    error += 1
        if error:
            return False
        return True

    def lun_query_access(self, name):
        """Query access information for a LUN"""
        if not name:
            _print("FAIL: usage lun_query_access(name)")
            return None

        if not self._lun_exist(name):
            _print("FAIL: lun_query_access() - LUN %s do NOT exist" % name)
            return None

        # cmds = ["volume select %s" % (name), "access show"]
        # ret, output = self._run(cmds, return_output = True, verbose = False)
        # if ret != 0:
        #    _print("FAIL: \"%s\" command" % ", ".join(cmds))
        #    print output
        #    return None
        #
        # access_dict = eqlparse_show.parse_show(output)
        # if "_access_records_" not in access_dict:
        #    return None
        # return access_dict["_access_records_"]
        info_dict = self.lun_info(name)
        if not info_dict:
            _print("FAIL: lun_query_access() - Could not query info for %s" % name)
            return None
        if "map_infos" in info_dict:
            return info_dict["map_infos"]
        return None

    def _lun_change_status(self, lun_name, status):
        """Have to run volume select lun_name and then set the status"""
        if not lun_name or not status:
            _print("FAIL: _lun_change_status() - requires lun_name and status parameter")
            return False
        # Before I thought I need to run volume select first and then set status,
        # but it seems it can be done in a single command line
        cmds = [f"volume select {lun_name} {status}"]
        ret, output = self._run(cmds, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: "%s" command' % ", ".join(cmds))
            print(output)
            return False
        # Need to update info of our volume
        # self.volume_dict[name] = self.lun_info(name)
        return True



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

# Initial version was developed by jdmorts, and it is available at:
# http://pydoc.net/Python/ail/0.2.8/ail.platforms.equallogic.parse_show/

__author__ = "jdmorts"
# Modified by: Bruno Goncalves <bgoncalv@redhat.com>

import re

from libsan import _print


def remove_col_wrap(column_spec, column_names, rows):
    """Given a set of rows and column widths, concatentate the fields of each
    column until the number of fields equals the number of columns. If two rows
    each contain five fields in five columns, this function returns one row of
    five fields in five columns. Any leading/trailing spaces are removed.

    This::

       a    b         c  d e
     ['----|---------|--|-|----------',
      ' ---|  -------|--|-| -----------']

    becomes::

     {'a': '-------',
      'b': '-------',
      'c': '----------------',
      'd': '----',
      'e': '--',
      'f': '---------------------'}

    :param column_spec: row widths stored as int
    :type column_spec: list of int
    :param column_names: used as keys to build record dict
    :type column_names: list of str
    :param rows: rows of a column-wrapped record
    """
    record = {}
    offset = 0

    for col_width, col_name in zip(column_spec, column_names):
        column = ""
        # Strip line ends and left justify so len(row) == sum(column_spec)
        for row in (r.rstrip("\n\r").ljust(sum(column_spec)) for r in rows):
            # Concatenate lines in the current column
            col_slice = slice(offset, offset + col_width)
            column += row[col_slice].strip()
        # Add column to record and set offset for next column
        record[col_name] = column
        offset += col_width + 1

    return record


def parse_show(output):
    """Parse Dell EqualLogic 'show' command output. While the formatting is
    consistent, it's tricksy on account of an ill-conceived column wrap. The
    following is sample output illustrating this curious 'feature'::

        EQGroup01> volume select MBX01-DB01 access show
        ID  Initiator                     Ipaddress       AuthMethod UserName   Apply-To
        --- ----------------------------- --------------- ---------- ---------- --------
        1   iqn.1998-01.com.vmware:xxxxxv *.*.*.*         none                  volume
              m01-4c08cb5e
        2   iqn.1998-01.com.vmware:xxxxxv *.*.*.*         none                  volume
              m02-7b27a2a8
        . . .

    Because the use of hyphens to denote column width is consistent, we can
    precisely specify the number of characters in each column using the
    number of hyphens observed. Also, notice the two-space indent where the
    'Initiator' column wraps; this must be stripped before concatenation.
    :param output: output from an EqualLogic CLI 'show' command
    output: list of str
    """
    col_spec = []  # Column slices
    col_names = []  # Column headers
    previous_line = ""  # Used to catch column headers once we see hypens
    saw_prompt = False
    records = {}

    try:
        # lines = deque(output)
        lines = output.split("\n")
        # print "DEBUG: output : %s" % output

    except TypeError as exc:
        raise type(exc)(f"Unexpected output type {type(output):s}") from exc  # noqa: TRY003

    line_id = 0
    sub_item = None
    while True:
        try:
            cur_line = lines[line_id].rstrip("\n")
            line_id += 1
            if not cur_line:
                continue

        except IndexError:
            break
        # Identify EQ prompts; ignore remaining output if we see two of them
        if cur_line.split(" ", 1)[0].endswith(">"):
            if saw_prompt:
                break
            saw_prompt = True
        elif cur_line.startswith("_"):
            # The setting from lines starting with _ will be stored with diferent key
            # for example volume select <volume> access show
            entry = cur_line.replace("_", "")
            entry = entry.replace(" ", "_")
            entry = entry.lower()
            sub_item = entry
            col_spec = []
            col_names = []
        elif cur_line.startswith("-"):
            col_spec = [len(c) for c in cur_line.split()]
            col_names = [col.lower() for col in previous_line.replace("-", "_").split()]
        # Once we know the column widths,
        elif col_spec:
            rows = [cur_line]
            # Gather continuing lines resulting from column wrap
            while True:
                try:
                    # If next line is a continuation of current line,
                    if lines[line_id].startswith(" "):
                        rows.append(lines[line_id])
                        line_id += 1
                    else:
                        break

                except IndexError:
                    break
            # Combine like columns in rows of the same record
            record = remove_col_wrap(col_spec, col_names, rows)
            # Store record in container object and use first column as dict key
            key = record[col_names[0]]
            if sub_item:
                if sub_item not in records:
                    records[sub_item] = {}
                records[sub_item][key] = record
            else:
                records[key] = record

        # For example Volume Information of specific LUN
        elif re.match(r"(.*): (\S+)", cur_line):
            m = re.match(r"(.*): (\S+)", cur_line)
            if not sub_item:
                _print("FAIL: parse_show() - Does not know to which sub item this info belongs")
                previous_line = cur_line
                continue
            key = m.group(1)
            key = key.replace(" ", "_")
            key = key.lower()
            value = m.group(2)
            next_line = lines[line_id].rstrip("\n")
            #: means new entry
            # starting with _ mean end of registry
            if not re.search(": ", next_line) and not re.match("_", next_line):
                value += next_line
            if sub_item not in records:
                records[sub_item] = {}
            records[sub_item][key] = value

        # Update previous_line to current line
        previous_line = cur_line

    return records



# Copyright (C) 2023 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""nvme.py: Module to manipulate NVME devices."""

__author__ = "Filip Suba"
__copyright__ = "Copyright (c) 2023 Red Hat, Inc. All rights reserved."

from os import listdir
from re import match

from libsan import _print
from libsan.host.cmdline import run
from libsan.host.linux import get_boot_device, get_device_wwid
from libsan.host.lvm import pv_query
from libsan.host.md import md_get_storage_dev, md_query
from libsan.host.mp import is_multipathd_running, multipath_query_all
from libsan.misc.size import size_bytes_2_size_human


def is_nvme_device(device: str) -> bool:
    """
    Checks if device is nvme device.
    """
    return bool(match("^nvme[0-9]n[0-9]$", device))


def get_nvme_device_names() -> list:
    """Return list of nvme devices.

    Returns:
    list: Return list of nvme devices
    """
    return [name for name in listdir("/sys/block") if is_nvme_device(name)]


def get_logical_block_size(nvme_device: str) -> str:
    cmd = f"cat /sys/block/{nvme_device}/queue/logical_block_size"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: get_logical_block_size() - Could not get logical block size for nvme device: {nvme_device}")
        print(output)
        return ""
    if not output:
        return ""
    return output


def get_physical_block_size(nvme_device: str) -> str:
    cmd = f"cat /sys/block/{nvme_device}/queue/physical_block_size"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: get_physical_block_size() - Could not get physical block size for nvme device: {nvme_device}")
        print(output)
        return ""
    if not output:
        return ""
    return output


def size_of_device(nvme_device: str) -> int:
    """
    Usage
        size_of_device(device)
    Purpose
        Given an nvme_device name. Eg. nvme0n1
    Parameter
        nvme_device
    Returns
        size in bytes

    """

    logical_block_size = get_logical_block_size(nvme_device)

    if not logical_block_size:
        return 0

    cmd = f"cat /sys/block/{nvme_device}/size"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: size_of_device() - Could not get sector size for device {nvme_device}")
        print(output)
        return 0
    if not output:
        return 0

    sector_size = output

    return int(logical_block_size) * int(sector_size)


def get_nvme_wwid(nvme_device: str) -> str:
    """
    Usage
        get_nvme_wwid(nvme_device)
    Purpose
        Given an NVMe device name. Eg. nvme0n1
    Parameter
        nvme_device   device to get wwid for
    Returns
        wwid:       eg. 360fff19abdd9f5fb943525d45126ca27
    """
    cmd = f"cat /sys/block/{nvme_device}/wwid"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: get_nvme_wwid() - Could not get wwid of nvme device: {nvme_device}")
        print(output)
        return ""
    if not output:
        return ""
    return output


def get_nvme_nqn(nvme_device: str) -> str:
    cmd = f"cat /sys/block/{nvme_device}/device/subsysnqn"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: get_nvme_nqn() - Could not get nqn of nvme device: {nvme_device}")
        print(output)
        return ""
    if not output:
        return ""
    return output


def get_nvme_uuid(nvme_device: str) -> str:
    cmd = f"cat /sys/block/{nvme_device}/uuid"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"WARN: get_nvme_uuid() - Could not get uuid of nvme device: {nvme_device}")
        print(output)
        return ""
    if not output:
        return ""
    return output


def get_nvme_state(nvme_device: str) -> str:
    cmd = f"cat /sys/block/{nvme_device}/device/state"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: get_nvme_subsystem_state() - Could not get state of nvme device: {nvme_device}")
        print(output)
        return ""
    if not output:
        return ""
    return output


def get_nvme_model(nvme_device: str) -> str:
    cmd = f"cat /sys/block/{nvme_device}/device/model"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: get_nvme_model() - Could not get model of nvme device: {nvme_device}")
        print(output)
        return ""
    if not output:
        return ""
    return output


def get_nvme_transport(nvme_device: str) -> str:
    cmd = f"cat /sys/block/{nvme_device}/device/transport"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: get_nvme_transport() - Could not get transport of nvme device: {nvme_device}")
        print(output)
        return ""
    if not output:
        return ""
    return output


def query_all_nvme_devices(nvme_device="") -> dict:
    """
    Query information of all NVMe devices and return them as a dict
    where NVMe device name is the dict key.
    If an NVMe device is given as argument, return its info
    Parameter:
    \tnvme_device (option):        NVMe device name. eg: 'nvme0n1'
    """
    nvme_devices = {}
    for device in get_nvme_device_names():
        if nvme_device and nvme_device != device:
            # optmization in case we requested specific device, do not query all
            continue
        nvme_wwid = get_nvme_wwid(device)
        nvme_uuid = get_nvme_uuid(device)
        nvme_nqn = get_nvme_nqn(device)
        size_bytes = size_of_device(device)
        logical_block_size = get_logical_block_size(device)
        physical_block_size = get_physical_block_size(device)
        nvme_model = get_nvme_model(device)
        state = get_nvme_state(device)
        transport = get_nvme_transport(device)
        nvme_info = {
            "name": device,
            "wwid": nvme_wwid,
            "uuid": nvme_uuid,
            "nqn": nvme_nqn,  # Uses scsi_id to query WWN
            "size": size_bytes,
            "size_human": size_bytes_2_size_human(size_bytes),
            "logical_block_size": logical_block_size,
            "physical_block_size": physical_block_size,
            "state": state,
            "model": nvme_model,
            "transport": transport,
        }
        nvme_devices[device] = nvme_info

    return nvme_devices


def get_free_nvme_devices(
    exclude_boot_device=True,
    exclude_lvm_device=True,
    exclude_mpath_device=True,
    exclude_md_device=True,
    filter_only=None,
) -> dict:
    all_nvme_devices = query_all_nvme_devices()
    if not all_nvme_devices:
        # could not find any nvme devices
        return {}

    pvs = pv_query()
    md_devices = md_query()
    boot_dev = get_boot_device()
    boot_wwid = None
    # if for some reason we boot from a single device, but this device is part of multipath device
    # the mpath device should be skipped as well
    if boot_dev:
        boot_wwid = get_device_wwid(boot_dev)

    all_mp_info = None
    if (is_multipathd_running()) and exclude_mpath_device:
        all_mp_info = multipath_query_all()
        if all_mp_info and "by_wwid" not in list(all_mp_info.keys()):
            # Fail querying mpath, setting it back to None
            all_mp_info = None

    chosen_devices = {}
    for nvme_device in list(all_nvme_devices.keys()):
        nvme_info = all_nvme_devices[nvme_device]
        # Skip if mpath device is used for boot
        if boot_wwid == nvme_info["wwid"] and exclude_boot_device:
            _print(f"DEBUG: get_free_nvme_devices() - skip {nvme_info['name']} as it is used for boot")
            continue

        # Skip if device is used by multipath
        if all_mp_info and nvme_info["wwid"] in list(all_mp_info["by_wwid"].keys()) and exclude_mpath_device:
            _print(f"DEBUG: get_free_nvme_devices() - skip {nvme_info['name']} as it is used for mpath")
            continue

        # Skip if it is used by Soft RAID
        if md_devices and exclude_md_device:
            used_by_md = False
            for md_dev in md_devices:
                storage_devs = md_get_storage_dev(md_dev)
                if not storage_devs:
                    continue
                for dev in storage_devs:
                    dev_wwid = get_device_wwid(dev)
                    if not dev_wwid:
                        continue
                    if dev_wwid == nvme_info["wwid"]:
                        _print(f"DEBUG: get_free_nvme_devices() - skip {nvme_info['name']} as it is used for md")
                        used_by_md = True
                        continue
            if used_by_md:
                continue

        # Skip if filter_only is specified
        filtered = False
        if filter_only is not None:
            for key in filter_only:
                if nvme_info[key] != filter_only[key]:
                    _print(
                        f"DEBUG: get_free_nvme_devices() - "
                        f"filtered {nvme_info['name']} as {key} is not {filter_only[key]}"
                    )
                    filtered = True
                    continue
        if filtered:
            continue

        chosen_devices[nvme_info["name"]] = nvme_info

        # Skip if it is used by LVM
        if pvs and exclude_lvm_device:
            for pv in pvs:
                if "/" + nvme_device in pv:
                    _print(f"DEBUG: get_free_nvme_devices() - skip {nvme_info['name']} as it is used for LVM")
                    chosen_devices.pop(nvme_info["name"])

    return chosen_devices



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""ssh.py: Module to handle ssh session."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import socket
import time

from ssh2.exceptions import AuthenticationError, SocketDisconnectError, SocketRecvError, Timeout
from ssh2.session import Session
from ssh2.utils import wait_socket

from libsan import _print

LIBSSH2_ERROR_EAGAIN = -37


def _check_timeout_value(timeout):
    # Do not overflow C long int
    return timeout if timeout <= 2**32 else 2**32


def _connect_socket(host, port, timeout=5, try_number=1):
    # This function is to handle SocketDisconnectError.
    # When we get it, we need to start the whole socket operation again.
    def _retry(_host, _port, _timeout, _try_number):
        time.sleep(5)
        return _connect_socket(_host, _port, _timeout, _try_number + 1)

    if try_number == 11:
        _print("FAIL: Could not establish socket connection after 10 tries.")
        return None
    try:
        # Create socket connection
        sock = socket.create_connection((host, port), timeout)
        # Create a session
        s = Session()
        s.handshake(sock)
    except SocketDisconnectError:
        _print(
            "DEBUG: Got SocketDisconnectError, trying to connect to socket again after 5 seconds. %s/100" % try_number
        )
        return _retry(host, port, timeout, try_number)
    except OSError as e:
        print(f"DEBUG: Got socket.error: '{e}', retrying connection to socket after 5 seconds. {try_number}/100")
        return _retry(host, port, timeout, try_number)
    return s


def connect(host, port=22, user=None, passwd=None, max_attempt=5):
    """Connect to a host using ssh
    The arguments are:
    \thost:                         Hostname
    \tport:                         Port number used to connect
    \tuser:                         username
    \tpasswd:                       Password
    \tmax_attempt                   Maximum attempt to connect (default:5)
    Returns:
    \tssh:                          ssh session
    \tNone:                         If there was some problem
    """

    # paramiko works this way
    if host == "":
        return None
    elif host is None:
        host = ""

    if not user:
        _print('FAIL: connect() needs "user" parameter')
        return None
    if not passwd:
        _print('FAIL: connect() needs "passwd" parameter')
        return None

    i = 0

    s = _connect_socket(host, port)

    if not s:
        return None

    while True:
        # Try to authenticate user
        try:
            s.userauth_password(user, passwd)
            break
        except AuthenticationError:
            print("Authentication failed when connecting to %s" % host)
            return None
        except (ValueError, OSError):
            print("Could not SSH to %s, waiting for it to start" % host)
            i += 1
        except SocketDisconnectError:
            print("Socket got disconnected in between, connecting again.")
            s = _connect_socket(host, port)
            i += 1
        except Exception as e:
            print("Could not SSH to %s" % host)
            print("Exception: %s" % e)
            return None
        # If we could not connect within set number of tries
        if i == max_attempt:
            print("Could not connect to %s. Giving up" % host)
            return None
        else:
            # Wait before next attempt
            time.sleep(2)
    return s


def disconnect(ssh_session):
    """Disconnect from a ssh session
    The arguments are:
    \tssh_session:                  ssh session, it is return by connect()
    Returns:
    \tTrue:                         Always
    """
    ssh_session.disconnect()
    return True


def read_chan(sock, ssh_session, chan, buff_size=1024, timeout=None, stderr=False):
    # Read channel output in non blocking way
    timeout = timeout or 2**32
    func = chan.read
    if stderr:
        func = chan.read_stderr
    size, tmp_buf = func(buff_size)
    time_end = time.time() + timeout
    while size == LIBSSH2_ERROR_EAGAIN:
        wait_socket(sock, ssh_session, timeout=timeout)
        size, tmp_buf = func(buff_size)
        if time.time() > time_end:
            raise Timeout
    return size, tmp_buf


def _execute_function(sock, ssh_session, func, **kwargs):
    # This executes function when the socket is not blocked (LIBSSH2_ERROR_EAGAIN)
    ret = func(**kwargs)
    while ret == LIBSSH2_ERROR_EAGAIN:
        wait_socket(sock, ssh_session)
        ret = func(**kwargs)
    return ret


def _execute(chan, command):
    # Wrapper for chan.execute to take keyword argument
    return chan.execute(command)


def run_cmd(
    ssh_session,
    cmd,
    verbose=True,
    return_output=False,
    invoke_shell=False,
    timeout=None,
    cr="\n",
    expect=None,
    command_expect=None,
):
    """Run a command to a specific ssh session
    The arguments are:
    \tssh_session:                  ssh session, it is returned by connect()
    \\tcmd:                         command to be executed
    \\tinvoke_shell:                run command in interactive shell
    \\texpect:                      if invoke_shell is true we try to read until find expect pattern
    \\tcommand_expect:              if multiple commands are sent with shell and expect, this is the patter to wait for
    Returns:
    \texit status:                  command status code, always 0 with interactive shell
    \tNone:                         If there was some problem
    """
    error = 0
    if timeout:
        # paramiko required timeout in secs, ssh2-python in milisecs
        timeout = _check_timeout_value(int(timeout * 1000))
    command_expect = command_expect or expect

    # This is for nonblocking mode, which allows to wait for socket to be ready
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    ssh_session.set_blocking(False)

    # by default we use exec_command to run a command as it has better control
    # when reading the buffers and supports exit code
    if not invoke_shell:
        for c in cmd if not isinstance(cmd, str) else [cmd]:
            chan = _execute_function(sock, ssh_session, ssh_session.open_session)
            if not chan:
                _print("FAIL: Could not create a chan")
                if return_output:
                    return 127, None
                return 127

            if timeout:
                ssh_session.set_timeout(timeout)

            if verbose:
                _print("INFO: Running ssh command '%s'" % c)
            try:
                _execute_function(sock, ssh_session, _execute, chan=chan, command=c)
            except Timeout:
                _print("FAIL: ssh - Got timeout (%ds) while executing command: '%s'" % (timeout, c))
                _execute_function(sock, ssh_session, chan.close)
                if return_output:
                    return 127, None
                return 127
            except Exception as e:
                _print("FAIL: ssh - Could not execute command: '%s'" % c)
                print("Failed due: %s" % repr(e))
                _execute_function(sock, ssh_session, chan.close)
                if return_output:
                    return 127, None
                return 127

            stdout = ""

            # keep reading stdout as long as there is anything to read
            try:
                size = 1
                while size > 0:
                    size, stdout_buf = read_chan(sock, ssh_session, chan, buff_size=1024, timeout=timeout)
                    stdout += stdout_buf.decode("ascii", "ignore")

                # Read stderr and add it to stdout
                size = 1
                while size > 0:
                    size, stderr_buff = read_chan(sock, ssh_session, chan, buff_size=1024, timeout=timeout, stderr=True)
                    stdout += stderr_buff.decode("ascii", "ignore")

                # The channel has to be closed to be able to get exit status
                _execute_function(sock, ssh_session, chan.close)

            except SocketRecvError:
                # Socket was closed, this happens when for example we reboot the server by this command
                pass

            exit_status = _execute_function(sock, ssh_session, chan.get_exit_status)

            if exit_status != 0:
                error += 1

            if verbose:
                print(stdout)

    else:
        # We use invoke_shell in situation where we need to run commands from a single shell
        # for example on cisco we need to run commands after configure terminal

        # Open channel
        chan = _execute_function(sock, ssh_session, ssh_session.open_session)

        # Set timeout
        if not timeout:
            timeout = 60 * 1000
        ssh_session.set_timeout(timeout)

        # Make this interactive shell
        _execute_function(sock, ssh_session, chan.pty)
        _execute_function(sock, ssh_session, chan.shell)

        tmp_buf = ""
        stdout = ""
        if expect:
            # wait for prompt
            while expect not in tmp_buf:
                try:
                    _, resp = read_chan(sock, ssh_session, chan, buff_size=9999, timeout=timeout)
                except Timeout:
                    _print(f"FAIL: Waiting for prompt '{expect}' timed out after {timeout} seconds.")
                    resp = expect
                tmp_buf += resp.decode("ascii", "ignore")

        else:
            # get rid of the whole header
            _execute_function(sock, ssh_session, chan.read, size=65535)

        # Handle cmd being a single string or a list of strings
        for c in cmd if not isinstance(cmd, str) else [cmd]:
            if verbose:
                print('INFO: ssh - sending command "%s"' % c)
            # Some targets need different ending symbols, for example ApCon switch needs \r, eqlogic \n\r...
            chan.write(c + cr)

            # Reading in between serves as waiting, as we are sending more commands when expecting prompt
            if expect:
                # On last command wait for normal prompt
                if c == cmd[-1]:
                    command_expect = expect
                # read until we get prompt again
                while not stdout.endswith(command_expect):
                    try:
                        size, tmp_buf = read_chan(sock, ssh_session, chan, buff_size=1024, timeout=timeout)
                    except Timeout:
                        _print(
                            "FAIL: Reading STDOUT timed out after %s seconds, did not get prompt '%s'."
                            % (timeout, command_expect)
                        )
                        tmp_buf = command_expect
                    stdout += tmp_buf.decode("ascii", "ignore")
                # Add newline at the end to prevent the above while condition to be True before reading some data
                stdout += "\n"

        if not expect:
            # Read until we do not receive more bytes. First STDOUT, than STDERR
            size = 1
            while size > 0:
                try:
                    size, tmp_buf = read_chan(sock, ssh_session, chan, buff_size=1024, timeout=timeout)
                except Timeout:
                    _print("FAIL: Reading STDOUT ssh channel timed out after %s seconds." % timeout)
                    size, tmp_buf = (0, "")
                stdout += tmp_buf.decode("ascii", "ignore")
            size = 1
            while size > 0:
                try:
                    size, tmp_buf = read_chan(sock, ssh_session, chan, buff_size=1024, timeout=timeout, stderr=True)
                except Timeout:
                    _print("WARN: Reading STDERR ssh channel timed out after %s seconds." % timeout)
                    size, tmp_buf = (0, "")
                stdout += tmp_buf.decode("ascii", "ignore")

        # Close channel at the end
        chan.close()

        error = 0
        if verbose:
            print(stdout)

    if return_output:
        return error, stdout

    return error
